//
// File:    hm_Args.c++
// Author:  Marc Zinnschlag

#ifndef HDR_hm_Args
#include "hm_Args.h"
#endif

// C

#ifndef LIB_cassert
#include <cassert>
#define LIB_cassert
#endif

// C++

#ifndef LIB_iostream
#include <iostream>
#define LIB_iostream
#endif

// HMake

#ifndef HDR_hm_Host
#include "hm_Host.h"
#endif

// -----------------------------------------------------------------------------------------
// Klasse: hm_args
// -----------------------------------------------------------------------------------------

// Konstruktoren ---------------------------------------------------------------------------

hm_args::hm_args (int argc, char *argv[])
: d_Options (0), d_Args (0), d_Ok (true)
{
    if (argc>1)
    {
        d_Args = 1;

#if HOST==RISC_OS
        d_Cpp = "c++";
#else
        d_Cpp = "cpp";
        d_LayoutPath = ".layout";
#endif

        for (int i=1; i<argc; ++i)
        {
            if (argv[i][0]=='-')
            {
                switch (argv[i][1])
                {
                    case 0: break;

                    case 'a': d_Options |= PointerAssert; break;
                    case 'h': d_Options |= HMake; break;
                    case 'c': d_Options |= CMake; break;
                    case 'v': d_Options |= PureVirtual; break;
                    case 'm': d_Options |= MakeMake; break;

                    case 'e':

                        if (argv[i][2])
                            d_Cpp = argv[i] + 2;
                        else
                        {
                            cerr << "-e incomplete" << endl;
                            d_Ok = false;
                        }

                        break;

                    case 'l':

                        if (argv[i][2])
                            d_LayoutPath = argv[i] + 2;
                        else
                        {
                            cerr << "-l incomplete" << endl;
                            d_Ok = false;
                        }

                        break;

                    default:
                        d_Ok = false;
                        cerr << "Option " << argv[i] << " unknown" << endl;
                }
            }
            else
            {
                if (d_SpecPath.empty())
                    d_SpecPath = argv[i];
                else
                    cerr << "Too many arguments" << endl;
            }
        }

        if (d_SpecPath.empty())
        {
            d_Ok = false;
            cerr << "You must specify a spec-file" << endl;
        }

        if (d_LayoutPath.empty())
        {
            d_Ok = false;
            cerr << "You must specify a layout-file" << endl;
        }

        if ((d_Options & (HMake | CMake | MakeMake))==0)
            d_Options |= HMake | CMake | MakeMake;
    }
}

hm_args::~hm_args() {}

// Konstante -------------------------------------------------------------------------------

void hm_args::ExplainOptions (ostream& Stream, const string& ToolName) const
{
    Stream
        << endl

        << "Syntax:   " << ToolName << " [options] spec-file" << endl

        << "options:" << endl << endl

//         << "-a        check for 0-pointer with assert" << endl
        << "-h        create h-files" << endl
        << "-c        create c-files" << endl
//         << "-v        don't delete pure virtual functions, when creating c-file" << endl
        << "-m        create make-file" << endl
        << "-eDotExt  set dot extension for c-file" << endl
        << "-lLayout  specify layout-file" << endl
        << endl;
}
