//
// File:    hm_Attribute.h
// Author:  Marc Zinnschlag

// Klassen:
// hm_output_protocol:
//      Interface fr Ausgabe durch Attribut-Klassen
// hm_attribute_protocol:
//      Basisklasse fr Attribute

#ifndef HDR_hm_Attribute
#define HDR_hm_Attribute

// C++

#ifndef LIB_string2
#include <string>
#define LIB_string2
#endif

#ifndef LIB_list
#include <list>
#define LIB_list
#endif

// -----------------------------------------------------------------------------------------
// Klasse: hm_output_protocol
// -----------------------------------------------------------------------------------------

using namespace std;

class hm_output_protocol
{
    public:

    // Konstruktoren

        hm_output_protocol();
        virtual ~hm_output_protocol();

    // Nicht-Konstante

        virtual void WriteFunction (const string& String) = 0;

        virtual void WriteComment (const string& String) = 0;

        virtual void NewFunction() = 0;
        // Alle weiteren Write-Aufrufe gehren zu einer anderen Funktion; mu nur aufgerufen
        // werden, wenn innerhalb eines Durchlaufs mehrere Funktionen erzeugt werden
};

// -----------------------------------------------------------------------------------------
// Klasse: hm_attribute_protocol
// -----------------------------------------------------------------------------------------

class hm_class;

class hm_attribute_protocol
{
    private:

    // Nicht implementiert

        hm_attribute_protocol (const hm_attribute_protocol&);
        hm_attribute_protocol& operator= (const hm_attribute_protocol&);

    public:

    // Konstruktoren

        hm_attribute_protocol();
        virtual ~hm_attribute_protocol();

    // Konstante

        virtual void WriteAttributes (list<string>& Attributes) const;
        // Bekannte Attribute eintragen

        virtual void WriteIncludes (const hm_class& Class, list<string>& Includes) const;
        // Zu inkludende h-files eintragen

        virtual void WriteFriends (const hm_class& Class, hm_output_protocol& Output) const;
        // Friend-Funktionen

        virtual void WriteNotImp (const hm_class& Class, hm_output_protocol& Output) const;
        // Nicht zu implementierende Funktionen (z.B. Copy-Konstruktor oder Zuweisung)

        virtual void WriteFree (const hm_class& Class, hm_output_protocol& Output) const;
        // Freie Funktionen

        virtual void WriteInline (const hm_class& Class, hm_output_protocol& Output) const;
        // inline-Funktionen

        virtual void WritePubCon (const hm_class& Class, hm_output_protocol& Output) const;
        // Konstruktoren; public

        virtual void WriteProCon (const hm_class& Class, hm_output_protocol& Output) const;
        // Konstruktoren; protected

        virtual void WritePriCon (const hm_class& Class, hm_output_protocol& Output) const;
        // Konstruktoren; private

        virtual void WritePubObs (const hm_class& Class, hm_output_protocol& Output) const;
        // Konstante Funktionen; public

        virtual void WriteProObs (const hm_class& Class, hm_output_protocol& Output) const;
        // Konstante Funktionen; protected

        virtual void WritePriObs (const hm_class& Class, hm_output_protocol& Output) const;
        // Konstante Funktionen; private

        virtual void WritePubMani (const hm_class& Class, hm_output_protocol& Output) const;
        // Nicht-Konstante Funktionen; public

        virtual void WriteProMani (const hm_class& Class, hm_output_protocol& Output) const;
        // Nicht-Konstante Funktionen; protected

        virtual void WritePriMani (const hm_class& Class, hm_output_protocol& Output) const;
        // Nicht-Konstante Funktionen; private

        virtual void WritePubStat (const hm_class& Class, hm_output_protocol& Output) const;
        // Statische Funktionen; public

        virtual void WriteProStat (const hm_class& Class, hm_output_protocol& Output) const;
        // Statische Funktionen; protected

        virtual void WritePriStat (const hm_class& Class, hm_output_protocol& Output) const;
        // Statische Funktionen; private
};

#endif
