//
// File:    hm_CMake.c++
// Author:  Marc Zinnschlag

#ifndef HDR_hm_CMake
#include "hm_CMake.h"
#endif

#ifndef HDR_hm_Host
#include "hm_Host.h"
#endif

// C++

#ifndef LIB_iostream
#include <iostream>
#define LIB_iostream
#endif

#ifndef LIB_fstream
#include <fstream>
#define LIB_fstream
#endif

#ifdef HAS_SSTREAM
#ifndef LIB_sstream
#include <sstream>
#define LIB_sstream
#endif
#else
#ifndef LIB_strstream
#include <strstream>
#define LIB_strstream
#endif
#endif

#ifndef LIB_string2
#include <string>
#define LIB_string2
#endif

#ifndef LIB_algorithm
#include <algorithm>
#define LIB_algorithm
#endif

#ifndef HDR_ul_Scanner
#include "ul_Scanner.h"
#endif

#ifndef HDR_ul_Token
#include "ul_Token.h"
#endif

// HMake

#ifndef HDR_hm_ClassMake
#include "hm_ClassMake.h"
#endif

#ifndef HDR_hm_Component
#include "hm_Component.h"
#endif

#ifndef HDR_hm_Functions
#include "hm_Functions.h"
#endif

// -----------------------------------------------------------------------------------------
// Klasse: hm_cmake
// -----------------------------------------------------------------------------------------

// Private Hilfsfunktionen -----------------------------------------------------------------

void hm_cmake::SkipInstruction (ul_scanner& HFile) const
{
    ul_token Token;
    while ((HFile >> Token) && Token!=ul_token (ul_token::Char, ';'));
    if (HFile) HFile << Token;
}

void hm_cmake::SkipTitle (ul_scanner& HFile, ostream& CFile) const
{
    HFile.EnablePreprocessor();

    ul_token Token;

    while (HFile >> Token &&
        (Token.GetType()==ul_token::CComment || Token.GetType()==ul_token::CppComment ||
        Token.GetType()==ul_token::Whitespace))
        CFile << Token;

    if (HFile) HFile << Token;

    HFile.DisablePreprocessor();
}

void hm_cmake::ReadTemplate (ul_scanner& HFile, list<ul_token>& Template) const
{
    int BCount = 0;
    ul_token Token;

    while ((HFile >> Token))
    {
        if (Token==ul_token (ul_token::Char, '>'))
        {
            Template.push_back (Token);
            break;
        }
        else if (Token==ul_token (ul_token::Char, '='))
        {
            // Default-Werte berspringen
            int BCount = 0;
            bool Continue = true;

            while (Continue && HFile >> Token)
                if (Token.GetType()==ul_token::Char)
                    switch (Token.GetChar())
                    {
                        case ',':

                            if (BCount==0) Continue = false;
                            break;

                        case '<': ++BCount; break;

                        case '>':

                            if (--BCount==0)
                            {
                                Continue = false;
                                HFile << Token;
                            }

                            break;
                    }
        }
        else
            Template.push_back (Token);
    }
}

// Return: Ok?

bool hm_cmake::MakeComponent (const list<hm_component>& Components,
    const hm_component& Component,
    const list<hm_attribute_protocol *>& Attributes)
{
#if HOST==RISCOS
    string HPath ("h.");
    string CPath (d_Args.GetCpp());
    CPath += '.';
    HPath += Component.GetName();
    CPath += Component.GetName();
#else
    string HPath (Component.GetName());
    string CPath (Component.GetName());
    HPath += ".h";
    CPath += ".";
    CPath += d_Args.GetCpp();
#endif

    // erster Pa (inline-Funktionen auflisten)
    {
#ifdef HAS_SSTREAM
        ostringstream Dummy;
#else
	ostrstream Dummy;
#endif

        list<ul_token> Template;

        ifstream HFile2 (HPath.c_str());
        ul_scanner HFile (HFile2);

        int Count = 0;
        map<string, string> Map;
        H2C (Dummy, HFile, Count, "", Map, Template);
    }

    // zweiter Pa
    list<ul_token> Template;
    ifstream HFile2 (HPath.c_str());
    ul_scanner HFile (HFile2);

    ofstream CFile (CPath.c_str());

    SkipTitle (HFile, CFile);

    // Includes
    if (!Component.HasMain())
    	CFile << "#include \"" << Component.GetName() << ".h\"" << endl;
	
    Component.WriteIncludes (CFile, false, Components);
    CFile << endl;

    // Klassen in C-File
    for (list<hm_class>::const_iterator Iter (Component.GetClasses (false).begin());
        Iter!=Component.GetClasses (false).end(); ++Iter)
    {
#ifdef HAS_SSTREAM
        ostringstream Temp;
#else
	ostrstream Temp;
#endif

        hm_class_make ClassMake (d_Layout, *Iter, Attributes);
        ClassMake.Write (Temp);

        CFile << Temp.str();

#ifdef HAS_SSTREAM
        istringstream Temp2 (Temp.str());
#else
	istrstream Temp2 (Temp.str());
#endif

        int Count = 0;
        map<string, string> Map;

        ul_scanner HFile (Temp2);
        H2C (CFile, HFile, Count, "", Map, Template);
    }

    // Rest
    int Count = 0;
    map<string, string> Map;
    H2C (CFile, HFile, Count, "", Map, Template);

    // Main
    if (Component.HasMain())
    {
        CFile << endl << "int main (int argc, char *argv[])";

        switch (d_Layout.GetIndentStyle())
        {
            case 0:

                CFile << " {" << endl << endl << '}' << endl;
                break;

            case 1:

                CFile << endl << '{' << endl << endl << '}' << endl;
                break;

            case 2:

                CFile << endl << d_Layout.GetIndent (hm_layout::Imp) << '{' << endl << endl
                      << d_Layout.GetIndent (hm_layout::Imp) << '}' << endl;
                break;
        }
    }

    return CFile.good();
}

// Types: typedefs und enums (first) mit Klassenangaben (second)

void hm_cmake::H2C (ostream& Out, ul_scanner& In, int& NestingCount, const string& Class,
    map<string, string> Types, const list<ul_token>& Template)
{
    int BaseCount = NestingCount;
    ul_token Token;
    bool Ignore = false;

    list<ul_token> LocalTemplate;

    while (In >> Token)
    {
        if (Token==ul_token (ul_token::Name, "template"))
        {
            In << Token;
            ReadTemplate (In, LocalTemplate);
        }
        else if (Token==ul_token (ul_token::Name, "enum"))
        {
            In.Disable (ul_token::CComment);
            In.Disable (ul_token::CppComment);
            In.Disable (ul_token::Whitespace);

            if (In >> Token)
            {
                if (Token.GetType()==ul_token::Name)
                {
                    if (!Class.empty())
                        Types.insert (make_pair (Token.GetString(), Class));
                    SkipInstruction (In);
                }
                else
                    In << Token;
            }

            In.Enable (ul_token::CComment);
            In.Enable (ul_token::CppComment);
            In.Enable (ul_token::Whitespace);
        }
        else if (Token==ul_token (ul_token::Name, "typedef"))
        {
            string Name;

            for (; In; In >> Token)
            {
                if (Token.GetType()==ul_token::Name)
                    Name = Token.GetString();
                else if (Token==ul_token (ul_token::Char, ';'))
                    break;
            }

            if (!Name.empty() && !Class.empty())
                Types.insert (make_pair (Name, Class));
        }
        else if (Token==ul_token (ul_token::Name, "struct") ||
            Token==ul_token (ul_token::Name, "class"))
        {
            In.Disable (ul_token::CComment);
            In.Disable (ul_token::CppComment);
            In.Disable (ul_token::Whitespace);

            if (In >> Token)
            {
                if (Token.GetType()==ul_token::Name)
                {
                    if (Class.empty())
                    {
                        d_Layout.WriteHeading (Token.GetString(), Out, hm_layout::Classes);
                        Out << endl;
                    }

                    map<string, string> Types2 (Types);

                    string Name (Token.GetString());

                    while (In >> Token)
                        if (Token==ul_token (ul_token::Char, '{') ||
                            Token==ul_token (ul_token::Char, ';'))
                            break;

                    In.Enable (ul_token::CComment);
                    In.Enable (ul_token::CppComment);
                    In.Enable (ul_token::Whitespace);

                    if (In && Token==ul_token (ul_token::Char, '{'))
                    {
                        In << Token;

                        string Class2 (Class);
                        if (!Class2.empty()) Class2 += "::";
                        Class2 += Name;

                        list<ul_token> Template2 (LocalTemplate);
                        Template2.insert (Template2.begin(), Template.begin(),
                            Template.end());

                        H2C (Out, In, NestingCount, Class2, Types2, Template2);

                        Ignore = false;
                    }
                }
                else
                    In << Token;
            }

            In.Enable (ul_token::CComment);
            In.Enable (ul_token::CppComment);
            In.Enable (ul_token::Whitespace);

            LocalTemplate.clear();
        }
        else if (Token==ul_token (ul_token::Char, '{'))
            ++NestingCount;
        else if (Token==ul_token (ul_token::Char, '}'))
        {
            if (--NestingCount<=0 && !Class.empty())
                return; // Ende einer Klasse
        }
        else if (Token==ul_token (ul_token::Name, "public") ||
            Token==ul_token (ul_token::Name, "private") ||
            Token==ul_token (ul_token::Name, "protected"))
        {
            In.Disable (ul_token::CComment);
            In.Disable (ul_token::CppComment);
            In.Disable (ul_token::Whitespace);

            if ((In >> Token) && Token!=ul_token (ul_token::Char, ':'))
                In << Token;

            In.Enable (ul_token::CComment);
            In.Enable (ul_token::CppComment);
            In.Enable (ul_token::Whitespace);

            Ignore = false;
        }
        else
        {
            // Auf Funktion prfen
            In << Token;

            list<ul_token> Template2 (LocalTemplate);
            Template2.insert (Template2.begin(), Template.begin(), Template.end());

            hm_function Function (d_Layout, In, Types, Class, Template2);

            if (Function.IsNIMark())
                Ignore = true;
            else if (!Function.IsEmpty())
            {
                if (!Ignore)
                {
                    if (Function.IsInline())
                    {
                        // In Auflistung der Inline-Funktionen aufnehmen
                        d_Inline.push_back (Function);
                    }
                    else
                    {
                        // Prfen, ob Funktion an einer anderen Stelle inline definiert ist
                        if (find (d_Inline.begin(), d_Inline.end(), Function)==
                            d_Inline.end())
                            Out << Function << endl;
                    }
                }

                LocalTemplate.clear();
            }
        }
    }
}

void hm_cmake::WriteTokens (const list<ul_token>::const_iterator& Begin,
    const list<ul_token>::const_iterator& End, ostream& CFile)
{
    for (list<ul_token>::const_iterator Iter (Begin); Iter!=End; ++Iter)
        CFile << *Iter;
}

// Konstruktoren ---------------------------------------------------------------------------

hm_cmake::hm_cmake (const hm_layout& Layout, const hm_args& Args)
: d_Layout (Layout), d_Args (Args)
{}

hm_cmake::~hm_cmake() {}

// Nicht-Konstante -------------------------------------------------------------------------

// Return: Ok?

bool hm_cmake::Make (const list<hm_component>& Components,
    const list<hm_attribute_protocol *>& Attributes)
{
    int Errors = 0;

    for (list<hm_component>::const_iterator Iter (Components.begin());
        Iter!=Components.end(); ++Iter)
        if (!MakeComponent (Components, *Iter, Attributes))
        {
            cerr << "Couldn't create " << (*Iter).GetName() << "." << d_Args.GetCpp()
                 << endl;

            ++Errors;
        }

    if (Errors)
    {
        cerr << "Aborted with " << Errors << " error";
        if (Errors>1) cerr << 's';
        cerr << '.' << endl;
    }

    return Errors==0;
}
