//
// File:    hm_ClassMake.c++
// Author:  Marc Zinnschlag

#ifndef HDR_hm_ClassMake
#include "hm_ClassMake.h"
#endif

// C

#ifndef LIB_cassert
#include <cassert>
#define LIB_cassert
#endif

// C++

#ifndef LIB_string2
#include <string>
#define LIB_string2
#endif

#ifndef LIB_list
#include <list>
#define LIB_list
#endif

#ifndef LIB_algorithm
#include <algorithm>
#define LIB_algorithm
#endif

#ifndef LIB_iostream
#include <iostream>
#define LIB_iostream
#endif

// HMake

#ifndef HDR_hm_Attribute
#include "hm_Attribute.h"
#endif

#ifndef HDR_hm_Class
#include "hm_Class.h"
#endif

#ifndef HDR_hm_Layout
#include "hm_Layout.h"
#endif

// -----------------------------------------------------------------------------------------
// Klasse: hm_function2
// -----------------------------------------------------------------------------------------

class hm_function2
{
        list<string> d_Function;
        list<string> d_Comments;

    private:

    // Konstante

        void WriteSingleFunction (ostream& Stream, int Width, const string& Indent,
            const string& String, int SubWidth, const string& SubIndent) const;

        void WriteSingleComment (ostream& Stream, int Width, const string& Indent,
            const string& String) const;

    public:

    // Konstruktoren

        hm_function2();
        ~hm_function2();

    // Konstante

        void WriteToStream (ostream& Stream, const hm_layout& Layout, bool Outside) const;

    // Nicht-Konstante

        void AddFunction (const string& String);

        void AddComment (const string& String);
};

// Private Hilfsfunktionen -----------------------------------------------------------------

void hm_function2::WriteSingleFunction (ostream& Stream, int Width, const string& Indent,
    const string& String, int SubWidth, const string& SubIndent) const
{
//     if (Width==-1 || String.length()<=Width)
        Stream << Indent << String << endl;
//     else
//     {
//
//     }
}

void hm_function2::WriteSingleComment (ostream& Stream, int Width, const string& Indent,
    const string& String) const
{
//     if (Width==-1 || String.length()<=Width-3)
        Stream << Indent << "// " << String << endl;
//     else
//     {
//
//     }
}

// Konstruktoren ---------------------------------------------------------------------------

hm_function2::hm_function2() {}

hm_function2::~hm_function2() {}

// Konstante -------------------------------------------------------------------------------

void hm_function2::WriteToStream (ostream& Stream, const hm_layout& Layout, bool Outside)
    const
{
    int Width = Layout.GetWidth();
    if (!Outside && Width!=-1)
    {
        Width -= Layout.GetIndentWidth (hm_layout::Member);
        if (Width<0) Width = 1;
    }

    string Indent;
    if (!Outside) Indent = Layout.GetIndent (hm_layout::Member);

    for (int i=0; i<3; ++i)
        if (i==1)
        {
            // Leerzeile zwischen Funktion und Kommentaren
            if (!d_Function.empty() && !d_Comments.empty() &&
                (Layout.GetFunctionCommentMode()==hm_layout::AMode ||
                Layout.GetFunctionCommentMode()==hm_layout::BMode))
                Stream << endl;
        }
        else
        {
            bool Function = (i==0)==
                (Layout.GetFunctionCommentMode()==hm_layout::aMode ||
                Layout.GetFunctionCommentMode()==hm_layout::AMode);

            if (Function)
            {
                if (!d_Function.empty())
                {
                    int SubWidth = Width;
                    if (SubWidth!=-1)
                    {
                        SubWidth -= Layout.GetIndentWidth (hm_layout::Sub);
                        if (SubWidth<0) SubWidth = 1;
                    }

                    string SubIndent (Indent);
                    SubIndent += Layout.GetIndent (hm_layout::Sub);

                    for (list<string>::const_iterator Iter (d_Function.begin());
                        Iter!=d_Function.end(); ++Iter)
                        WriteSingleFunction (Stream, Width, Indent, *Iter, SubWidth,
                            SubIndent);
                }
            }
            else
                if (!d_Comments.empty())
                    for (list<string>::const_iterator Iter (d_Comments.begin());
                        Iter!=d_Comments.end(); ++Iter)
                        WriteSingleComment (Stream, Width, Indent, *Iter);
        }

    Stream << endl;
}

// Nicht-Konstante -------------------------------------------------------------------------

void hm_function2::AddFunction (const string& String)
{
    d_Function.push_back (String);
}

void hm_function2::AddComment (const string& String)
{
    d_Comments.push_back (String);
}

// -----------------------------------------------------------------------------------------
// Klasse: hm_output
// -----------------------------------------------------------------------------------------

class hm_output : public hm_output_protocol
{
        hm_layout::access d_Access;
        char d_Id;
        list<hm_function2> d_Functions;
        bool d_New;

    public:

    // Konstruktoren

        hm_output (hm_layout::access Access, char Id);
        virtual ~hm_output();

    // Konstante

        hm_layout::access GetAccess() const;
        char GetId() const;

        bool IsEmpty() const;

        void WriteToStream (ostream& Stream, const hm_layout& Layout, bool Outside) const;

    // Nicht-Konstante

        virtual void WriteFunction (const string& String);

        virtual void WriteComment (const string& String);

        virtual void NewFunction();
        // Alle weiteren Write-Aufrufe gehren zu einer anderen Funktion; mu nur aufgerufen
        // werden, wenn innerhalb eines Durchlaufs mehrere Funktionen erzeugt werden
};

// Konstruktoren ---------------------------------------------------------------------------

hm_output::hm_output (hm_layout::access Access, char Id)
: d_Access (Access), d_Id (Id), d_New (true)
{}

hm_output::~hm_output() {}

// Konstante -------------------------------------------------------------------------------

inline hm_layout::access hm_output::GetAccess() const
{
    return d_Access;
}

inline char hm_output::GetId() const
{
    return d_Id;
}

inline bool hm_output::IsEmpty() const
{
    return d_Functions.empty();
}

void hm_output::WriteToStream (ostream& Stream, const hm_layout& Layout, bool Outside) const
{
    for (list<hm_function2>::const_iterator Iter (d_Functions.begin());
        Iter!=d_Functions.end(); ++Iter)
        (*Iter).WriteToStream (Stream, Layout, Outside);
}

// Nicht-Konstante -------------------------------------------------------------------------

void hm_output::WriteFunction (const string& String)
{
    if (!String.empty())
    {
        if (d_New)
        {
            d_New = false;
            d_Functions.push_back (hm_function2 ());
        }

        list<hm_function2>::iterator Iter (d_Functions.end());
        (*--Iter).AddFunction (String);
    }
}

void hm_output::WriteComment (const string& String)
{
    if (!String.empty())
    {
        if (d_New)
        {
            d_New = false;
            d_Functions.push_back (hm_function2 ());
        }

        list<hm_function2>::iterator Iter (d_Functions.end());
        (*--Iter).AddComment (String);
    }
}

// Alle weiteren Write-Aufrufe gehren zu einer anderen Funktion; mu nur aufgerufen
// werden, wenn innerhalb eines Durchlaufs mehrere Funktionen erzeugt werden

void hm_output::NewFunction()
{
    d_New = true;
}

// -----------------------------------------------------------------------------------------
// Klasse: hm_class_make
// -----------------------------------------------------------------------------------------

// Private Hilfsfunktionen -----------------------------------------------------------------

void hm_class_make::AddOutput (hm_layout::access Access, char Id, const hm_class& Class,
    const list<hm_attribute_protocol *>& Attributes, bool Outside)
{
    (Outside ? d_OutputOutside : d_Output).push_back (hm_output (Access, Id));
    list<hm_output>::iterator End ((Outside ? d_OutputOutside : d_Output).end());
    hm_output& Output (*--End);

    void (hm_attribute_protocol::*Func) (const hm_class&, hm_output_protocol&) const = 0;

    switch (Id)
    {
        case 'F': Func = &hm_attribute_protocol::WriteFriends; break;
        case 'n': Func = &hm_attribute_protocol::WriteNotImp; break;
        case 'f': Func = &hm_attribute_protocol::WriteFree; break;
        case 'i': Func = &hm_attribute_protocol::WriteInline; break;
        case 'C': Func = &hm_attribute_protocol::WritePubCon; break;
        case 'O': Func = &hm_attribute_protocol::WritePubObs; break;
        case 'M': Func = &hm_attribute_protocol::WritePubMani; break;
        case 'S': Func = &hm_attribute_protocol::WritePubStat; break;
        case 'c': Func = &hm_attribute_protocol::WritePriCon; break;
        case 'o': Func = &hm_attribute_protocol::WritePriObs; break;
        case 'm': Func = &hm_attribute_protocol::WritePriMani; break;
        case 's': Func = &hm_attribute_protocol::WritePriStat; break;
        case 'd': Func = &hm_attribute_protocol::WriteProCon; break;
        case 'p': Func = &hm_attribute_protocol::WriteProObs; break;
        case 'l': Func = &hm_attribute_protocol::WriteProMani; break;
        case 't': Func = &hm_attribute_protocol::WriteProStat; break;
    }

    assert (Func);

    for (list<hm_attribute_protocol *>::const_iterator Iter (Attributes.begin());
        Iter!=Attributes.end(); ++Iter)
    {
        (*(*Iter).*Func) (Class, Output); // <--- Sorry !!!
        Output.NewFunction();
    }
}

// Konstruktoren ---------------------------------------------------------------------------

hm_class_make::hm_class_make (const hm_layout& Layout, const hm_class& Class,
    const list<hm_attribute_protocol *>& Attributes)
: d_Layout (Layout), d_Class (Class.GetName())
{
    // Datenstruktur aufbauen
    int Len = d_Layout.GetOrder().length();
    for (int i=0; i<Len; ++i)
    {
        hm_layout::access Access = hm_layout::Public;
        char Id = d_Layout.GetOrder()[i];
        bool Outside = false;

        switch (Id)
        {
            case 'n': Access = hm_layout::Private; break;
            case 'c': Access = hm_layout::Private; break;
            case 'o': Access = hm_layout::Private; break;
            case 'm': Access = hm_layout::Private; break;
            case 's': Access = hm_layout::Private; break;
            case 'f': Outside = true; break;
            case 'd': Access = hm_layout::Protected; break;
            case 'p': Access = hm_layout::Protected; break;
            case 'l': Access = hm_layout::Protected; break;
            case 't': Access = hm_layout::Protected; break;
        }

        AddOutput (Access, Id, Class, Attributes, Outside);
    }

    AddOutput (hm_layout::Public, 'F', Class, Attributes, false); // friends
    AddOutput (hm_layout::Public, 'i', Class, Attributes, true); // inline
}

hm_class_make::~hm_class_make() {}

// Konstante -------------------------------------------------------------------------------

// Klasse in Stream schreiben

void hm_class_make::Write (ostream& Stream) const
{
    d_Layout.WriteHeading (d_Class, Stream, hm_layout::Classes);

    Stream << endl << "class " << d_Class << endl << '{' << endl;

    hm_layout::access Access = hm_layout::Private;

    for (list<hm_output>::const_iterator Iter (d_Output.begin()); Iter!=d_Output.end();
        ++Iter)
    {
        if ((*Iter).IsEmpty() &&
            (!(d_Layout.GetEmptyState ((*Iter).GetAccess())) || (*Iter).GetId()=='n'))
            continue;

        if ((*Iter).GetAccess()!=Access && (*Iter).GetId()!='F')
        {
            Access = (*Iter).GetAccess();
            Stream << d_Layout.GetIndent (hm_layout::Access);
            switch (Access)
            {
                case hm_layout::Public: Stream << "public:"; break;
                case hm_layout::Protected: Stream << "protected:"; break;
                case hm_layout::Private: Stream << "private:"; break;
            }

            Stream << endl << endl;
        }

        hm_layout::heading Heading;
        switch ((*Iter).GetId())
        {
            case 'c':
            case 'C':
	    case 'd':

                Heading = hm_layout::Con;
                break;

            case 'o':
            case 'O':
	    case 'p':

                Heading = hm_layout::Obs;
                break;

            case 'm':
            case 'M':
	    case 'l':

                Heading = hm_layout::Mani;
                break;

            case 's':
            case 'S':
	    case 't':

                Heading = hm_layout::Stat;
                break;

            case 'n':

                Heading = hm_layout::NotImp;
                break;

            case 'F':

                Heading = hm_layout::Friends;
                break;
        }

        if (!d_Layout.GetHeading (Heading).empty())
        {
            d_Layout.WriteHeading (d_Class, Stream, Heading);
            Stream << endl;
        }

        (*Iter).WriteToStream (Stream, d_Layout, false);
    }

    Stream << "};" << endl;

    for (list<hm_output>::const_iterator Iter (d_OutputOutside.begin());
        Iter!=d_OutputOutside.end(); ++Iter)
        if (!(*Iter).IsEmpty())
            switch ((*Iter).GetId())
            {
                case 'f':

                    Stream << endl;
                    d_Layout.WriteHeading (d_Class, Stream, hm_layout::Free);
                    Stream << endl;
                    (*Iter).WriteToStream (Stream, d_Layout, true);
                    break;

                case 'i':

                    Stream << endl;
                    (*Iter).WriteToStream (Stream, d_Layout, true);
                    break;
            }

    Stream << endl;
}
