//
// File:    hm_Component.c++
// Author:  Marc Zinnschlag

#ifndef HDR_hm_Component
#include "hm_Component.h"
#endif

// C

#ifndef LIB_cassert
#include <cassert>
#define LIB_cassert
#endif

// C++

#ifndef LIB_algorithm
#include <algorithm>
#define LIB_algorithm
#endif

#ifndef LIB_iostream
#include <iostream>
#define LIB_iostream
#endif

// -----------------------------------------------------------------------------------------
// Klasse: hm_component
// -----------------------------------------------------------------------------------------

// Private Hilfsfunktionen -----------------------------------------------------------------

// Include-Eintrge in Klasseneintrge umwandeln, aus UnusedClasses herausnehmen und
// in UsedClasses eintragen
// Return: Ok? (falls nicht wird ein passende Fehlermeldung ausgegeben)

bool hm_component::MakeClasses (list<hm_class>& UnusedClasses, list<string>& UsedClasses,
    list<hm_class>& Classes, list<string>& Include)
{
    bool Good = true;

    // Include-Liste durchlaufen und nach Eintrgen suchen, die gar keine Includes sondern
    // Klassen sind
    for (list<string>::iterator Iter (Include.begin()); Iter!=Include.end(); ++Iter)
    {
        hm_class Value (*Iter);

        // Unbenutzte Klassen
        list<hm_class>::iterator Result =
            find (UnusedClasses.begin(), UnusedClasses.end(), Value);

        if (Result==UnusedClasses.end())
        {
            // Benutzte Klassen berprfen
            list<string>::iterator Result =
                find (UsedClasses.begin(), UsedClasses.end(), *Iter);
            if (Result!=UsedClasses.end())
            {
                Good = false;
                cerr << "Error (component " << d_Name << "): Class " << *Result
                     << " is used twice (second appearance ignored)" << endl;
            }
            else
                continue; // Include ist tatschlich Include
        }
        else
        {
            // Include ist kein Include, sondern Klasse -> bertragen
            UsedClasses.push_back ((*Result).GetName());
            Classes.push_back (*Result);
            UnusedClasses.erase (Result);
        }

        // Include ist kein Include oder fehlderhaftes Include -> aus Include-Liste
        //  entfernen
        list<string>::iterator Iter2 (Iter);
        --Iter;
        Include.erase (Iter2);
    }

    return Good;
}

// Konstruktoren ---------------------------------------------------------------------------

hm_component::hm_component (const string& Name) : d_Name (Name), d_Main (false) {}

hm_component::~hm_component() {}

// Konstante -------------------------------------------------------------------------------

// include-Anweisungen schreiben

void hm_component::WriteIncludes (ostream& Stream, bool HFile,
    const list<hm_component>& Components) const
{
    const list<string>& Include = HFile ? d_HInclude : d_CInclude;

    for (list<string>::const_iterator Iter (Include.begin()); Iter!=Include.end(); ++Iter)
    {
        const string& Include (*Iter);

        assert (!Include.empty());
        string Pre, Post;
        if (Include[0]!='"' && Include[0]!='<')
        {
            if (Include[Include.length()-1]=='h')
                Pre = Post = "\"";
            else
            {
                list<hm_component>::const_iterator Iter (Components.begin());
                for (; Iter!=Components.end(); ++Iter)
                    if ((*Iter).GetName()==Include)
                        break;

                if (Iter!=Components.end())
                {
                    Pre = "\"";
                    Post = ".h\"";
                }
                else
                {
                    Pre = "<";
                    Post = ">";
                }
            }
        }

        Stream << "#include " << Pre << Include << Post << endl;
    }

    if (!Include.empty()) Stream << endl;
}

// Nicht-Konstante -------------------------------------------------------------------------

// Argument aus Spec-Datei eintragen
// Ausgenommen von speziellen Argumenten (z.B. main) werden alle Argumente zunchst
// als Include-Anweisungen interpretiert; die Klassenzuordnung mu spter mit MakeClasses
// erfolgen
// Arg darf kein leerer String sein

void hm_component::AddArg (const string& Arg, bool HFile)
{
    assert (!Arg.empty());

    if (Arg=="main")
    {
        if (HFile)
            cerr << "Warning (component " << d_Name
                 << "): Shouldn't use 'main' in h-file (treated as used in c-file)"
                 << endl;

        d_Main = true;
    }
    else
        (HFile ? d_HInclude : d_CInclude).push_back (Arg);
}

// Include-Eintrge in Klasseneintrge umwandeln, aus UnusedClasses herausnehmen und in
// UsedClasses eintragen
// Return: Ok? (falls nicht wird ein passende Fehlermeldung ausgegeben)

bool hm_component::MakeClasses (list<hm_class>& UnusedClasses, list<string>& UsedClasses)
{
    bool Good = MakeClasses (UnusedClasses, UsedClasses, d_HClasses, d_HInclude);

    if (!MakeClasses (UnusedClasses, UsedClasses, d_CClasses, d_CInclude))
        Good = false;

    return Good;
}

// Doppelte Include-Anweisungen entfernen

void hm_component::CleanIncludes()
{
    unique (d_HInclude.begin(), d_HInclude.end());
    unique (d_CInclude.begin(), d_CInclude.end());
}

// Statische -------------------------------------------------------------------------------

// Include-Dateinamen von <> bzw. "" und .h befreien

string hm_component::StripInclude (const string& Include)
{
    if (!Include.empty())
    {
        int Begin = 0;
        int End = Include.length();
        if (Begin<End && (Include[0]=='<' || Include[0]=='"'))
            ++Begin;

        if (Begin<End && (Include[End-1]=='>' || Include[End-1]=='"'))
            --End;

        if (Begin<End-1 && (Include[End-1]=='h' || Include[End-1]=='H') &&
            Include[End-2]=='.')
            End -= 2;

        return Include.substr (Begin, End-Begin);
    }

    return Include;
}
