//
// File:    hm_Component.h
// Author:  Marc Zinnschlag

// Klassen:
// hm_component:
//      Komponente (h- und c-Datei)

#ifndef HDR_hm_Component
#define HDR_hm_Component

// C++

#ifndef LIB_string2
#include <string>
#define LIB_string2
#endif

#ifndef LIB_list
#include <list>
#define LIB_list
#endif

// HMake

#ifndef HDR_hm_Class
#include "hm_Class.h"
#endif

// -----------------------------------------------------------------------------------------
// Klasse: hm_component
// -----------------------------------------------------------------------------------------

using namespace std;

#include <iosfwd>

class hm_class;

class hm_component
{
        string d_Name;
        list<hm_class> d_HClasses;
        list<hm_class> d_CClasses;
        list<string> d_HInclude;
        list<string> d_CInclude;
        bool d_Main;

    private:

    // Nicht-Konstante

        bool MakeClasses (list<hm_class>& UnusedClasses, list<string>& UsedClasses,
            list<hm_class>& Classes, list<string>& Include);
        // Include-Eintrge in Klasseneintrge umwandeln, aus UnusedClasses herausnehmen und
        // in UsedClasses eintragen
        // Return: Ok? (falls nicht wird ein passende Fehlermeldung ausgegeben)

    public:

    // Konstruktoren

        hm_component (const string& Name);
        ~hm_component();

    // Konstante

        string GetName() const;

        const list<hm_class>& GetClasses (bool HFile) const;

        const list<string>& GetIncludes (bool HFile) const;

        bool HasMain() const;

        void WriteIncludes (ostream& Stream, bool HFile,
            const list<hm_component>& Components) const;
        // include-Anweisungen schreiben

    // Nicht-Konstante

        void AddArg (const string& Arg, bool HFile);
        // Argument aus Spec-Datei eintragen
        // Ausgenommen von speziellen Argumenten (z.B. main) werden alle Argumente zunchst
        // als Include-Anweisungen interpretiert; die Klassenzuordnung mu spter mit
        // MakeClasses erfolgen
        // Arg darf kein leerer String sein

        bool MakeClasses (list<hm_class>& UnusedClasses, list<string>& UsedClasses);
        // Include-Eintrge in Klasseneintrge umwandeln, aus UnusedClasses herausnehmen und
        // in UsedClasses eintragen
        // Return: Ok? (falls nicht wird ein passende Fehlermeldung ausgegeben)

        void CleanIncludes();
        // Doppelte Include-Anweisungen entfernen

    // Statische

        static string StripInclude (const string& Include);
        // Include-Dateinamen von <> bzw. "" und .h befreien
};

// Freie Operatoren

bool operator== (const hm_component& Left, const hm_component& Right);
bool operator!= (const hm_component& Left, const hm_component& Right);

// inline-Funktionen

inline string hm_component::GetName() const { return d_Name; }
inline const list<hm_class>& hm_component::GetClasses (bool HFile) const
{ return HFile ? d_HClasses : d_CClasses; }
inline const list<string>& hm_component::GetIncludes (bool HFile) const
{ return HFile ? d_HInclude : d_CInclude; }
inline bool hm_component::HasMain() const { return d_Main; }
inline bool operator== (const hm_component& Left, const hm_component& Right)
{ return Left.GetName()==Right.GetName(); }
inline bool operator!= (const hm_component& Left, const hm_component& Right)
{ return !(Left==Right); }

#endif
