//
// File:    hm_Components.c++
// Author:  Marc Zinnschlag

// Klassen:
// hm_components:
//      Komponenten-Struktur (aus Spec-File gelesen)

#ifndef HDR_hm_Components
#include "hm_Components.h"
#endif

// C

#ifndef LIB_cassert
#include <cassert>
#define LIB_cassert
#endif

// C++

#ifndef LIB_algorithm
#include <algorithm>
#define LIB_algorithm
#endif

#ifndef LIB_iostream
#include <iostream>
#define LIB_iostream
#endif

// HMake

#ifndef HDR_hm_SpecScan
#include "hm_SpecScan.h"
#endif

#ifndef HDR_hm_Class
#include "hm_Class.h"
#endif

// -----------------------------------------------------------------------------------------
// Klasse: hm_components
// -----------------------------------------------------------------------------------------

// Private Hilfsfunktionen -----------------------------------------------------------------

// Spec-Datei auswerten

void hm_components::Scan (istream& Stream, const list<string>& Attributes,
    list<hm_class>& Classes)
{
    // Einlesen
    hm_spec_scanner Scanner (Stream, Attributes);

    hm_spec_line Line;

    while (Scanner.Get (Line))
    {
        if (Line.GetType()==hm_spec_line::NoFile)
        {
            // Klasse
            hm_class Value (Line.GetFile());
            list<hm_class>::iterator Result = find (Classes.begin(), Classes.end(), Value);

            hm_class& Class = Result==Classes.end() ? Value : *Result;

            // Attribute bernehmen
            for (list<string>::const_iterator Iter (Line.GetArgs().begin());
                Iter!=Line.GetArgs().end(); ++Iter)
                Class.AddAttribute (*Iter);

            if (Result==Classes.end())
            {
                // Neue Klasse
                Classes.push_back (Value);
            }
        }
        else
        {
            // Komponente
            hm_component Value (Line.GetFile());
            list<hm_component>::iterator Result = find (d_Components.begin(),
                d_Components.end(), Value);

            hm_component& Component = Result==d_Components.end() ? Value : *Result;

            // Argumente bernehmen
            for (list<string>::const_iterator Iter (Line.GetArgs().begin());
                Iter!=Line.GetArgs().end(); ++Iter)
                Component.AddArg (*Iter, Line.GetType()==hm_spec_line::HFile);

            if (Result==d_Components.end())
            {
                // Neue Komponente
                d_Components.push_back (Value);
            }
        }
    }

    if (Scanner.HadErrors()) d_Errors = true;
}

// Klassenangaben in gelesener Spec-Datei identifizieren

void hm_components::IdentifyClasses (list<hm_class>& Classes)
{
    // Klassen identifizieren & aufrumen
    list<string> UsedClasses;
    for (list<hm_component>::iterator Iter (d_Components.begin()); Iter!=d_Components.end();
        ++Iter)
    {
        if (!(*Iter).MakeClasses (Classes, UsedClasses))
            d_Errors = true;

        (*Iter).CleanIncludes();
    }
}

// Auf ungenutzte Klassen prfen

void hm_components::CheckClasses (const list<hm_class>& Classes)
{
    // Ungenutzte Klassen berprfen
    if (!Classes.empty())
    {
        cerr << endl
             << "Warning: the following classes have been mentioned, but have no "
             << "specification:" << endl;
        for (list<hm_class>::const_iterator Iter (Classes.begin()); Iter!=Classes.end();
            ++Iter)
            cerr << (*Iter).GetName() << endl;

        cerr << endl;
    }
}

// Konstruktoren ---------------------------------------------------------------------------

// Stream: Spec-Datei

hm_components::hm_components (istream& Stream, const list<string>& Attributes)
: d_Errors (false)
{
    list<hm_class> Unused;

    Scan (Stream, Attributes, Unused);
    IdentifyClasses (Unused);
    CheckClasses (Unused);
}
