//
// File:    hm_Default.c++
// Author:  Marc Zinnschlag

#ifndef HDR_hm_Default
#include "hm_Default.h"
#endif

// C++

#ifndef LIB_string2
#include <string>
#define LIB_string2
#endif

#ifndef LIB_list
#include <list>
#define LIB_list
#endif

// HMake

#ifndef HDR_hm_Class
#include "hm_Class.h"
#endif

// -----------------------------------------------------------------------------------------
// Klasse: hm_default_attributes
// -----------------------------------------------------------------------------------------

static const char s_Ops[] = { '+', '-', '*', '/', 0 };

// Konstruktoren ---------------------------------------------------------------------------

hm_default_attributes::hm_default_attributes() {}

hm_default_attributes::~hm_default_attributes() {}

// Konstante -------------------------------------------------------------------------------

// Bekannte Attribute eintragen

void hm_default_attributes::WriteAttributes (list<string>& Attributes) const
{
    static const char *Names[] =
    {
        "copy", "nocopy", "assign", "noassign", "compare", "fullcompare", "stream",
        "op+", "op-", "op*", "op/", 0
    };

    for (int i=0; Names[i]; ++i)
        Attributes.push_back (Names[i]);
}

// Zu inkludende h-files eintragen

void hm_default_attributes::WriteIncludes (const hm_class& Class, list<string>& Includes)
    const
{
    if (Class.HasAttribute ("stream"))
        Includes.push_back ("iostream");
}

// Nicht zu implementierende Funktionen (z.B. Copy-Konstruktor oder Zuweisung)

void hm_default_attributes::WriteNotImp (const hm_class& Class, hm_output_protocol& Output)
    const
{
    if (Class.HasAttribute ("nocopy"))
    {
    	Output.NewFunction();
	
        string Copy (Class.GetName());
        Copy += " (const ";
        Copy += Class.GetName();
        Copy += "&);";
        Output.WriteFunction (Copy);
    }

    if (Class.HasAttribute ("noassign"))
    {
        Output.NewFunction();

        string Assign (Class.GetName());
        Assign += "& operator= (const ";
        Assign += Class.GetName();
        Assign += "&);";
        Output.WriteFunction (Assign);
    }
}

// Freie Funktionen

void hm_default_attributes::WriteFree (const hm_class& Class, hm_output_protocol& Output)
    const
{
    if (Class.HasAttribute ("compare") || Class.HasAttribute ("fullcompare"))
    {
        string Compare ("bool operator== (const ");
        Compare += Class.GetName();
        Compare += "& Left, const ";
        Compare += Class.GetName();
        Compare += "& Right);";
        Output.WriteFunction (Compare);

        Output.NewFunction();
        Compare[13] = '!';
        Output.WriteFunction (Compare);
    }

    if (Class.HasAttribute ("fullcompare"))
    {
        Output.NewFunction();

        string Compare ("bool operator>  (const ");
        Compare += Class.GetName();
        Compare += "& Left, const ";
        Compare += Class.GetName();
        Compare += "& Right);";
        Output.WriteFunction (Compare);

        Output.NewFunction();
        Compare[14] = '=';
        Output.WriteFunction (Compare);

        Output.NewFunction();
        Compare[13] = '<';
        Output.WriteFunction (Compare);

        Output.NewFunction();
        Compare[14] = ' ';
        Output.WriteFunction (Compare);
    }

    string Op ("opx");
    for (int i=0; s_Ops[i]; ++i)
    {
        Op[2] = s_Ops[i];
        if (Class.HasAttribute (Op))
        {
            Output.NewFunction();

            string String (Class.GetName());
            String += " operator";
            String += s_Ops[i];
            String += " (const ";
            String += Class.GetName();
            String += "& Left, const ";
            String += Class.GetName();
            String += "& Right);";
            Output.WriteFunction (String);
        }
    }
    
    if (Class.HasAttribute ("stream"))
    {
    	Output.NewFunction();
	
	string Out ("ostream");
	Out += "& operator<< (ostream& Stream, const ";
	Out += Class.GetName();
	Out += "& Object);";
	Output.WriteFunction (Out);
	
    	Output.NewFunction();
	
	string In ("istream");
	In += "& operator>> (istream& Stream, ";
	In += Class.GetName();
	In += "& Object);";
	Output.WriteFunction (In);
    }
}

// inline-Funktionen

void hm_default_attributes::WriteInline (const hm_class& Class, hm_output_protocol& Output)
    const
{
    if (Class.HasAttribute ("compare") || Class.HasAttribute ("fullcompare"))
    {
        string Compare ("inline bool operator!= (const ");
        Compare += Class.GetName();
        Compare += "& Left, const ";
        Compare += Class.GetName();
        Compare += "& Right) { return !(Left==Right); } ";
        Output.WriteFunction (Compare);
    }

    string Op ("opx");
    for (int i=0; s_Ops[i]; ++i)
    {
        Op[2] = s_Ops[i];
        if (Class.HasAttribute (Op))
        {
            Output.NewFunction();

            string String ("inline ");
            String += Class.GetName();
            String += " operator";
            String += s_Ops[i];
            String += " (const ";
            String += Class.GetName();
            String += "& Left, const ";
            String += Class.GetName();
            String += "& Right) { return ";
            String += Class.GetName();
            String += " (Left) ";
            String += s_Ops[i];
            String += "= Right; }";
            Output.WriteFunction (String);
        }
    }
}

// Konstruktoren; public

void hm_default_attributes::WritePubCon (const hm_class& Class, hm_output_protocol& Output)
    const
{
    if (Class.HasAttribute ("copy"))
    {
        string Copy (Class.GetName());
        Copy += " (const ";
        Copy += Class.GetName();
        Copy += "& Obj);";
        Output.WriteFunction (Copy);
    }
}

// Nicht-Konstante Funktionen; public

void hm_default_attributes::WritePubMani (const hm_class& Class, hm_output_protocol& Output)
    const
{
    if (Class.HasAttribute ("assign"))
    {
        string Assign (Class.GetName());
        Assign += "& operator= (const ";
        Assign += Class.GetName();
        Assign += "& Obj);";
        Output.WriteFunction (Assign);
    }

    string Op ("opx");
    for (int i=0; s_Ops[i]; ++i)
    {
        Op[2] = s_Ops[i];
        if (Class.HasAttribute (Op))
        {
            Output.NewFunction();

            string String (Class.GetName());
            String += "& operator";
            String += s_Ops[i];
            String += "= (const ";
            String += Class.GetName();
            String += "& Obj);";
            Output.WriteFunction (String);
        }
    }
}
