//
// File:    hm_Functions.c++
// Author:  Marc Zinnschlag

#ifndef HDR_hm_Functions
#include "hm_Functions.h"
#endif

// C

#ifndef LIB_ctype
#include <cctype>
#define LIB_ctype
#endif

// C++

#ifndef LIB_iostream
#include <iostream>
#define LIB_iostream
#endif

#ifndef LIB_algorithm
#include <algorithm>
#define LIB_algorithm
#endif

#ifndef LIB_functional
#include <functional>
#define LIB_funtional
#endif

#ifndef HDR_ul_Scanner
#include "ul_Scanner.h"
#endif

// HMake

#ifndef HDR_hm_Layout
#include "hm_Layout.h"
#endif

// -----------------------------------------------------------------------------------------
// Klasse: hm_function
// -----------------------------------------------------------------------------------------

// Private Hilfsfunktionen -----------------------------------------------------------------

// Unres Prdikat: Whitespace, aber kein Newline

static bool IsNonNLWhitespace (const ul_token& Token)
{
    return Token.GetType()==ul_token::Whitespace && Token.GetChar()!='\n';
}

// Unres Prdikat: !(Whitespace, aber kein Newline)

static bool IsNotNoneNLWhitespace (const ul_token& Token)
{
    return !(Token.GetType()==ul_token::Whitespace && Token.GetChar()!='\n');
}

static bool IsEqualIgnoreCase (char Left, char Right)
{
    return tolower (Left)==tolower (Right);
}

void hm_function::WriteTokens (const list<ul_token>& Tokens, ostream& Stream) const
{
    for (list<ul_token>::const_iterator Iter (Tokens.begin()); Iter!=Tokens.end(); ++Iter)
        Stream << *Iter;
}

// Kommentare auf NotImplementet berprfen

bool hm_function::NotImplementedTest() const
{
    typedef list<ul_token>::const_iterator t_token_citer;

    string NotImplemented (d_Layout.GetHeading (hm_layout::NotImp));

    for (t_token_citer Iter (d_Comments.begin()); Iter!=d_Comments.end(); ++Iter)
        if (Iter->GetType()==ul_token::CComment || Iter->GetType()==ul_token::CppComment)
        {
            string String (Iter->GetString());

            if (search (String.begin(), String.end(), NotImplemented.begin(),
                NotImplemented.end(), IsEqualIgnoreCase)!=String.end())
                return true;
        }

    return false;
}

// EmptyLine: Sich anschlieende Leerzeile?
// NoPreEmptyLine: Keine vorangestellte Leerzeile zulassen
// Indent: In: Erforderliche Einrcktiefe (-1: egal, -2: egal, aber mu einheitlich sein),
//         Out: tatschliche Einrcktiefe
// Return: kann Kommentar zu folgender Funktion gehren?

bool hm_function::ReadComments (ul_scanner& Scanner, bool& EmptyLine, bool NoPreEmptyLine,
    int& Indent)
{
    typedef list<ul_token>::reverse_iterator t_token_riter;
    typedef list<ul_token>::iterator t_token_iter;

    EmptyLine = false;
    int EmptyLineCount = 0;
    bool WhitespacesOnly = true; // bisher nur Whitespaces gelesen

    int Indent2 = 0; // Vorgefundene Einrcktiefe
    int IndentReturn = -1;
    bool ReadingIndent = true; // Gelesene Tokens gehren zur Einrckung

    ul_token Token;
    bool NewLine = true; // Vorangegangenes NewLine mit nur Whitespaces danach
    t_token_iter EmptyLinePos = d_Comments.end();

    while (Scanner >> Token)
    {
        if (Token.GetType()!=ul_token::Whitespace &&
            Token.GetType()!=ul_token::CComment &&
            Token.GetType()!=ul_token::CppComment)
        {
            // Ende des Kommentars
            Scanner << Token;
            break;
        }
        else if (Token.GetType()==ul_token::Whitespace)
        {
            if (Token.GetChar()=='\n')
            {
                if (NewLine && !d_Comments.empty())
                {
                    // zwei auf einander folgende NewLines mit ausschlielich Whitespaces
                    // dazwischen -> Leerzeile gefunden
                    if (NoPreEmptyLine && WhitespacesOnly)
                        break;

                    EmptyLine = true;
                    if (EmptyLineCount++)
                    {
                        // Zwei Leerzeilen -> Abbrechen
                        Scanner << Token;

                        list<ul_token> Back;
                        Back.splice (Back.begin(), d_Comments, ++EmptyLinePos,
                            d_Comments.end());

                        for (t_token_riter Iter (Back.rbegin()); Iter!=Back.rend(); ++Iter)
                            Scanner << *Iter;

                        break;
                    }
                    else
                        EmptyLinePos = --d_Comments.end();
                }

                // Indent2 mu hier nicht berprft werde, da Zeile nicht leer ist und somit
                // schon eine berprfung stattgefunden hat

                NewLine = ReadingIndent = true;
            }
            else if (ReadingIndent)
            {
                if (Token.GetChar()==' ')
                    ++Indent2;
                else if (Token.GetChar()=='\t')
                    Indent2 += 4 - Indent2%4;
            }
        }
        else
        {
            // Kommentar
            ReadingIndent = WhitespacesOnly = NewLine = false;

            // Einrcktiefe berprfen
            if (IndentReturn==-1) IndentReturn = Indent2;
            if (Indent==-2) Indent = Indent2;
            if (Indent!=-1 && Indent!=Indent2)
            {
                // berprfung gescheitert -> Zurckschreiben und Abbrechen
                Scanner << Token;

                t_token_riter Newline (find (d_Comments.rbegin(), d_Comments.rend(),
                    ul_token (ul_token::Whitespace, '\n')));

                list<ul_token> Back;

                Back.splice (Back.begin(), d_Comments, Newline.base(), d_Comments.end());

                for (t_token_riter Iter (Back.rbegin()); Iter!=Back.rend(); ++Iter)
                    Scanner << *Iter;

                EmptyLineCount = 2; // Kommentar kann nicht zu folgender Funktion gehren

                break;
            }
            else
                Indent2 = 0;
        }

        d_Comments.push_back (Token);
    }

    Indent = IndentReturn;

    if (!d_Comments.empty())
    {
        t_token_riter Newline = find (d_Comments.rbegin(), d_Comments.rend(),
            ul_token (ul_token::Whitespace, '\n'));

        // Tokens hinter letztem Newline entfernen und in Scanner zurcklegen
        list<ul_token> Back;

        Back.splice (Back.begin(), d_Comments, Newline.base(), d_Comments.end());

        for (t_token_riter Iter (Back.rbegin()); Iter!=Back.rend(); ++Iter)
            Scanner << *Iter;

        // Leere Kommentarzeilen entfernen, Einrckung entfernen
        t_token_iter Iter (d_Comments.begin());

        while (Iter!=d_Comments.end())
        {
            t_token_iter Iter2 (Iter);
            for (; Iter2!=d_Comments.end() && IsNonNLWhitespace (*Iter2); ++Iter2);

            Iter2 = Iter = d_Comments.erase (Iter, Iter2);

            // Hinter nchstes Zeilenende vorrcken
            Iter = find (Iter, d_Comments.end(), ul_token (ul_token::Whitespace, '\n'));

            if (Iter!=d_Comments.end())
            {
                if (Iter==Iter2)
                {
                    // Leerzeile
                    Iter = d_Comments.erase (Iter);
                }
                else
                    ++Iter; // Anfang nchster Zeile
            }
        }
    }

    return EmptyLineCount<=1;
}

// Indent: In: Erforderliche Einrcktiefe (-1: egal), Out: tatschliche Einrcktiefe
// der ersten Zeile
// Return: Funktion gelesen

bool hm_function::ReadFunction (ul_scanner& Scanner, int& Indent)
{
    typedef list<ul_token>::reverse_iterator t_token_riter;
    typedef list<ul_token>::iterator t_token_iter;

    ul_token Token;

    // Funktion lesen
    int OpenCount0 = 0; // ()
    int OpenCount1 = 0; // []
    int OpenCountT = -1; // <> (nur nach Template Anweisung)
    bool Ignore = false; // inline oder pure-virtual
    bool Found = false; // Funktion gefunden?
    bool End = false; // Ende der Deklaration erreicht
    bool TempIgnore = false; // Tokens vorrbergehend nicht bernehmen
    int TempCount0 = 0; // OpenCount0 beim Setzen von TempIgnore = true
    bool NonComment = false; // Zeichen gefunden, das nicht zu reiner Kommentarzeile gehren
                             // kann
    bool FirstLine = true;

    int Indent2 = 0; // Vorgefundene Einrcktiefe
    bool ReadingIndent = true; // Gelesene Tokens gehren zur Einrckung

    while (Scanner >> Token)
    {
        // Einrckung berprfen
        if (ReadingIndent)
        {
            if (Token==ul_token (ul_token::Whitespace, ' '))
                ++Indent2;
            else if (Token==ul_token (ul_token::Whitespace, '\t'))
                Indent2 += 4 - Indent2%4;
            else
            {
                ReadingIndent = false;
                if (Indent!=-1 && Indent2!=Indent)
                {
                    // Vorangegangener Kommentar stimmt in Einrcktiefe nicht mit
                    // Funktion berein -> Funktion zurck schreiben
                    Scanner << Token;
                    for (t_token_riter Iter (d_Tokens.rbegin()); Iter!=d_Tokens.rend();
                        ++Iter)
                        Scanner << *Iter;

                    d_Tokens.clear();
                    return false;
                }
                else
                    Indent = Indent2;

            }
        }

        if (Token==ul_token (ul_token::Char, ';') ||
            Token==ul_token (ul_token::Char, '}'))
            break;

        if (Token==ul_token (ul_token::Char, '{'))
        {
            NonComment = true;

            if (!Found) break;

            // inline-Implementierung
            int Open = 1;

            while (Open && (Scanner >> Token))
                if (Token.GetType()==ul_token::Char)
                    switch (Token.GetChar())
                    {
                        case '{': ++Open; break;
                        case '}': --Open; break;
                    }

            d_Tokens.push_front (ul_token (ul_token::Name, "inline"));
            End = true;
            break;
        }
        else
        {
            if (Token.GetType()==ul_token::Char)
            {
                NonComment = true;

                switch (Token.GetChar())
                {
                    case '(':

                        if (!OpenCount1)
                        {
                            ++OpenCount0;
                            if (!TempIgnore) Found = true; // Beginn der Parameterliste
                        }

                        break;

                    case ')':

                        if (!OpenCount1 && !--OpenCount0)
                        {
                            // Auf operator()() prfen
                            t_token_iter EndT (d_Tokens.end());
                            if (!(EndT!=d_Tokens.begin() && --EndT!=d_Tokens.begin() &&
                                *EndT==ul_token (ul_token::Name, "operator")))
                            {
                                // letzte Klammer geschlossen
                                TempIgnore = false;
                            }
                        }

                        break;

                    case '<':

                        if (OpenCountT>-1)
                            ++OpenCountT;

                        break;

                    case '>':

                        if (OpenCountT>0)
                            if (--OpenCountT==0)
                                OpenCountT = -1; // Ende des template-Konstukts

                        break;

                    case '[':

                        ++OpenCount1;
                        break;

                    case ']':

                        if (OpenCount1) --OpenCount1;
                        break;

                    case '=':
                    {
                        if (OpenCountT<=0)
                        {
                            // Auf operator= und Varianten prfen
                            t_token_iter EndT (d_Tokens.end());
                            if (EndT!=d_Tokens.begin())
                            {
                                if (*--EndT==ul_token (ul_token::Name, "operator"))
                                    break;
                                else if (EndT!=d_Tokens.begin() &&
                                    *--EndT==ul_token (ul_token::Name, "operator"))
                                    break;
                            }

                            if (!OpenCount0)
                            {
                                // rein virtuelle Funktion
                                Ignore = true;
                                End = true;
                            }
                            else
                            {
                                // Default-Parameter berlesen
                                TempIgnore = true;
                                TempCount0 = OpenCount0;

                                // Leerzeichen vor = entfernen
                                t_token_iter End (d_Tokens.end());
                                while (End!=d_Tokens.begin())
                                {
                                    --End;
                                    if (End->GetType()==ul_token::Whitespace ||
                                        End->GetType()==ul_token::CComment ||
                                        End->GetType()==ul_token::CppComment)
                                        End = d_Tokens.erase (End);
                                    else
                                        break;
                                }
                            }
                        }

                        break;
                    }

                    case ':':

                        if (!OpenCount0)
                        {
                            // Initialisierungsliste -> inline-Funktion
                            Ignore = true;
                            End = true;
                        }

                        break;

                    case ',':

                        if (TempIgnore && TempCount0==OpenCount0)
                            TempIgnore = false; // Ende eines Default-Wertes

                        break;
                }
            }
            else if (Token.GetType()==ul_token::Name)
            {
                NonComment = true;

                if (Token.GetString()=="const")
                {
                    if (Found) d_Const = true; // Konstante Funktion
                }
                else if (Token.GetString()=="class" || Token.GetString()=="struct" ||
                    Token.GetString()=="enum" || Token.GetString()=="typedef" ||
                    Token.GetString()=="public" || Token.GetString()=="private" ||
                    Token.GetString()=="protected")
                {
                    if (OpenCountT<=0)
                    {
                        Scanner << Token;
                        break;
                    }
                }
                else if (Token.GetString()=="template")
                {
                    OpenCountT = 0;
                }
            }
            else if (Token.GetType()==ul_token::Whitespace)
            {
                if (Token.GetChar()=='\n' && !NonComment && FirstLine)
                {
                    // Zeile ist doch nur Kommentar-Zeile -> zurckschreiben und abbrechen
                    Scanner << Token;

                    for (t_token_riter Iter (d_Tokens.rbegin()); Iter!=d_Tokens.rend();
                        ++Iter)
                        Scanner << *Iter;

                    d_Tokens.clear();

                    return false;
                }

                FirstLine = false;
            }
            else if (!(Token.GetType()==ul_token::CComment ||
                Token.GetType()==ul_token::CppComment))
                NonComment = true;

            if (!TempIgnore && !End)
                d_Tokens.push_back (Token);
        }
    }

    if (!Found) return false;

    // Folgende Whitespaces bis zum nchsten NewLine lesen
    while (Scanner >> Token)
    {
        if (Token.GetType()==ul_token::Whitespace)
        {
            if (Token.GetChar()=='\n')
                break;
        }
        else
        {
            Scanner << Token;
            break;
        }
    }

    // Auswertung und Bearbeitung der gefundenen Funktion
    if (Ignore)
    {
        // Funktion lschen
        d_Tokens.clear();
        // (Es wird trotzdem true zurckgegeben, um eventuelle nachfolgende Kommentare
        // zu lesen; da d_Tokens leer ist, wird die Funktion trotzdem ignoriert)
    }
    else
    {
        // Newline am letzten Zeilenende der Funktion entfernen
        t_token_iter End (d_Tokens.end());
        while (End!=d_Tokens.begin())
        {
            --End;
            if (End->GetType()==ul_token::Whitespace ||
                End->GetType()==ul_token::CComment ||
                End->GetType()==ul_token::CppComment)
                End = d_Tokens.erase (End);
            else
                break;
        }

        // inline-Erkennung
        const ul_token Inline (ul_token::Name, "inline");
        if (find (d_Tokens.begin(), d_Tokens.end(), Inline)!=d_Tokens.end())
            d_Inline = true;

        // explicit, virtual und static lschen
        const ul_token Static (ul_token::Name, "static");
        if (find (d_Tokens.begin(), d_Tokens.end(), Static)!=d_Tokens.end())
            d_Static = true;

        d_Tokens.remove (ul_token (ul_token::Name, "explicit"));
        d_Tokens.remove (ul_token (ul_token::Name, "virtual"));
        d_Tokens.remove (Inline);
        d_Tokens.remove (Static);

        // Leerzeichen am Zeilenanfang entfernen
        d_Tokens.erase (d_Tokens.begin(),
            find_if (d_Tokens.begin(), d_Tokens.end(), IsNotNoneNLWhitespace));
    }

    return true;
}

void hm_function::InsertScope (const string& Class)
{
    typedef list<ul_token>::iterator t_token_iter;

    if (!Class.empty())
    {
        t_token_iter Token =
            find (d_Tokens.begin(), d_Tokens.end(), ul_token (ul_token::Name, "operator"));

        if (Token!=d_Tokens.end())
        {
            // Operator
            d_Tokens.insert (Token, ul_token (ul_token::Name, Class));
            d_Tokens.insert (Token, ul_token (ul_token::Special, ul_token::Scope));
        }
        else
        {
            // Funktion
            Token =
                find (d_Tokens.begin(), d_Tokens.end(), ul_token (ul_token::Char, '('));
            if (Token!=d_Tokens.end() && Token!=d_Tokens.begin())
            {
                --Token;
                while (Token!=d_Tokens.begin() &&
                    (Token->GetType()==ul_token::Whitespace ||
                    Token->GetType()==ul_token::CComment ||
                    Token->GetType()==ul_token::CppComment))
                    --Token;

                d_Tokens.insert (Token, ul_token (ul_token::Name, Class));
                d_Tokens.insert (Token, ul_token (ul_token::Special, ul_token::Scope));
            }
        }
    }
}

void hm_function::InsertTypeScope (const map<string, string> Types)
{
    typedef list<ul_token>::iterator t_token_iter;
    typedef map<string, string>::const_iterator t_types_citer;

    for (t_types_citer Type (Types.begin()); Type!=Types.end(); ++Type)
    {
        ul_token Token (ul_token::Name, Type->first);

        t_token_iter Iter (d_Tokens.begin());

        while ((Iter = find (Iter, d_Tokens.end(), Token))!=d_Tokens.end())
        {
            t_token_iter Test (Iter);

            while (Test!=d_Tokens.begin())
            {
                --Test;
                if (IsRelevant (*Test))
                    break;
            }

            if (*Test!=ul_token (ul_token::Special, ul_token::Scope))
            {
                d_Tokens.insert (Iter, ul_token (ul_token::Name, Type->second));
                d_Tokens.insert (Iter, ul_token (ul_token::Special, ul_token::Scope));
            }

            ++Iter;
        }
    }
}

bool hm_function::IsRelevant (const ul_token& Token)
{
    return !(Token.GetType()==ul_token::CComment ||
        Token.GetType()==ul_token::CppComment ||
        Token.GetType()==ul_token::Whitespace);
}

// Konstruktoren ---------------------------------------------------------------------------

hm_function::hm_function (const hm_layout& Layout)
: d_Layout (Layout), d_Static (false), d_Const (false), d_NI (false), d_Inline (false)
{}

// Funktion mit Kommentaren einlesen
// Falls aus Scanner keine Funktion folgt, werden die gelesenen Tokens zurckgeschrieben
// Types: typedefs und enums (first) mit Klassenangaben (second)

hm_function::hm_function (const hm_layout& Layout, ul_scanner& Scanner,
  const map<string, string> Types, const string& Class, const list<ul_token>& Template)
: d_Layout (Layout), d_Class (Class), d_Static (false), d_Const (false), d_NI (false),
  d_Inline (false), d_Template (Template)
{
    typedef list<ul_token>::reverse_iterator t_token_riter;

    bool EmptyLine;
    int Indent = -1;
    if (Layout.HasStrictFunctionComments()) Indent = -2;
    if (!ReadComments (Scanner, EmptyLine, false, Indent))
        return;

    // Auf NI-Markierung prfen
    if (NotImplementedTest())
    {
        d_NI = true;
        return;
    }

    // Vorangestellte Kommentare lesen
    if (!(Layout.GetFunctionCommentMode()==hm_layout::bMode && !EmptyLine ||
        Layout.GetFunctionCommentMode()==hm_layout::BMode) &&
        !d_Comments.empty())
        return; // Nicht zu Funktion gehrende Kommenare gelesen (Kommentar verworfen)

    // Funktion lesen
    if (!Layout.HasStrictFunctionComments() || d_Comments.empty()) Indent = -1;
    if (!ReadFunction (Scanner, Indent))
    {
        // Keine Funktion gelesen (vorangestellter Kommentar bleibt zur Auswertung erhalten)
        d_Tokens = list<ul_token>();
        return;
    }

    InsertTypeScope (Types);
    if (!Class.empty()) InsertScope (Class);

    // Nachgestellte Kommentare lesen
    if (!Layout.HasStrictFunctionComments()) Indent = -1;
    if (Layout.GetFunctionCommentMode()==hm_layout::aMode ||
        Layout.GetFunctionCommentMode()==hm_layout::AMode)
    {
        ReadComments (Scanner, EmptyLine,
            Layout.GetFunctionCommentMode()==hm_layout::aMode, Indent);

        if (NotImplementedTest())
        {
            // Kommentare zurckschreiben
            for (t_token_riter Iter (d_Comments.rbegin()); Iter!=d_Comments.rend(); ++Iter)
                Scanner << *Iter;

            d_Comments.clear();
        }
    }
}

hm_function::~hm_function() {}

// Konstante -------------------------------------------------------------------------------

bool hm_function::IsEqual (const hm_function& Function) const
{
    typedef list<ul_token>::const_iterator t_token_citer;

    t_token_citer Iter2 (Function.d_Tokens.begin());

    for (t_token_citer Iter (d_Tokens.begin()); Iter!=d_Tokens.end(); ++Iter)
    {
        while (Iter2!=Function.d_Tokens.end() && !IsRelevant (*Iter2))
            ++Iter2;

        if (IsRelevant (*Iter))
        {
            if (Iter2==Function.d_Tokens.end()) return false;

            if (*Iter!=*Iter2) return false;
            ++Iter2;
        }
    }

    return Iter2==Function.d_Tokens.end();
}

// Funktion in C-File schreiben

void hm_function::Write (ostream& Stream) const
{
    // Kommentare
    if (!d_Comments.empty())
    {
        WriteTokens (d_Comments, Stream);
        Stream << endl;
    }

    // Template
    if (!d_Template.empty())
    {
        WriteTokens (d_Template, Stream);
        Stream << ' ';
    }

    // Functions-Head
    WriteTokens (d_Tokens, Stream);

    // Function-Body
    switch (d_Layout.GetIndentStyle())
    {
        case 1:

            Stream << " {" << endl << endl << '}' << endl;
            break;

        case 0:

            Stream << endl << '{' << endl << endl << '}' << endl;
            break;

        case 2:

            Stream << endl << d_Layout.GetIndent (hm_layout::Imp) << '{' << endl << endl
                   << d_Layout.GetIndent (hm_layout::Imp) << '}' << endl;
            break;
    }
}
