//
// File:    hm_Function.h
// Author:  Marc Zinnschlag

// Klassen:
// hm_function:
//      Funktionsdeklaration
//      Einlesen aus Scanner mit zugehrigen Kommentaren

#ifndef HDR_hm_Functions
#define HDR_hm_Functions

// C++

#ifndef LIB_string2
#include <string>
#define LIB_string2
#endif

#ifndef LIB_list
#include <list>
#define LIB_list
#endif

#ifndef LIB_map
#include <map>
#define LIB_map
#endif

#ifndef HDR_ul_Token
#include "ul_Token.h"
#endif

// HMake

#ifndef HDR_hm_Layout
#include "hm_Layout.h"
#endif

// -----------------------------------------------------------------------------------------
// Klasse: hm_function
// -----------------------------------------------------------------------------------------

using namespace std;

#include <iosfwd>

class ul_scanner;
class hm_layout;

class hm_function
{
        hm_layout d_Layout;
        list<ul_token> d_Tokens;
        list<ul_token> d_Comments;
        const list<ul_token> d_Template;
        string d_Class;
        bool d_Static;
        bool d_Const;
        bool d_NI;
        bool d_Inline;

    private:

    // Konstante

        void WriteTokens (const list<ul_token>& Tokens, ostream& Stream) const;

        bool NotImplementedTest() const;
        // Kommentare auf NotImplementet berprfen

    // Nicht-Konstante

        bool ReadComments (ul_scanner& Scanner, bool& EmptyLine, bool NoPreEmptyLine,
            int& Indent);
        // EmptyLine: Sich anschlieende Leerzeile?
        // NoPreEmptyLine: Keine vorangestellte Leerzeile zulassen
        // Indent: In: Erforderliche Einrcktiefe (-1: egal), Out: tatschliche Einrcktiefe
        // Return: kann Kommentar zu folgender Funktion gehren?

        bool ReadFunction (ul_scanner& Scanner, int& Indent);
        // Indent: In: Erforderliche Einrcktiefe (-1: egal), Out: tatschliche Einrcktiefe
        // der ersten Zeile
        // Return: Funktion gelesen?

        void InsertScope (const string& Class);

        void InsertTypeScope (const map<string, string> Types);

    // Statische

        static bool IsRelevant (const ul_token& Token);

    public:

    // Konstruktoren

        hm_function (const hm_layout& Layout);

        hm_function (const hm_layout& Layout, ul_scanner& Scanner,
            const map<string, string> Types, const string& Class = "",
            const list<ul_token>& Template = list<ul_token>());
        // Funktion mit Kommentaren einlesen
        // Falls aus Scanner keine Funktion folgt, werden die gelesenen Tokens
        // zurckgeschrieben
        // Types: typedefs und enums (first) mit Klassenangaben (second)

        ~hm_function();

    // Konstante

        bool IsEmpty() const;

        bool IsEqual (const hm_function& Function) const;

        void Write (ostream& Stream) const;
        // Funktion in C-File schreiben

        bool IsStatic() const;

        bool IsConst() const;

        bool IsNIMark() const;
        // NotImplemented-Markierung?

        bool IsInline() const;
        // Hat Funktion inline-Schlsselwort (tatschliches Schlsselwort wird entfernt)
};

// Freie Operatoren

bool operator== (const hm_function& Left, const hm_function& Right);
bool operator!= (const hm_function& Left, const hm_function& Right);

ostream& operator<< (ostream& Stream, const hm_function& Function);

// inline-Funktionen

inline bool hm_function::IsEmpty() const { return d_Tokens.empty(); }
inline bool hm_function::IsStatic() const { return d_Static; }
inline bool hm_function::IsConst() const { return d_Const; }
inline bool hm_function::IsNIMark() const { return d_NI; }
inline bool hm_function::IsInline() const { return d_Inline; }
inline bool operator== (const hm_function& Left, const hm_function& Right)
{ return Left.IsEqual (Right); }
inline bool operator!= (const hm_function& Left, const hm_function& Right)
{ return !(Left==Right); }
inline ostream& operator<< (ostream& Stream, const hm_function& Function)
{ Function.Write (Stream); return Stream; }

#endif
