//
// File:    hm_HCMake.c++
// Author:  Marc Zinnschlag

// C

#ifndef LIB_stdlib
#include <cstdlib>
#define LIB_stdlib
#endif

// C++

#ifndef LIB_iostream
#include <iostream>
#define LIB_iostream
#endif

#ifndef LIB_fstream
#include <fstream>
#define LIB_iostream
#endif

#ifndef LIB_new
#include <new>
#define LIB_new
#endif

#ifndef LIB_string2
#include <string>
#define LIB_string2
#endif

#ifndef LIB_list
#include <list>
#define LIB_list
#endif

#ifndef LIB_algorithm
#include <algorithm>
#define LIB_algorithm
#endif

// HMake

#ifndef HDR_hm_Args
#include "hm_Args.h"
#endif

#ifndef HDR_hm_Layout
#include "hm_Layout.h"
#endif

#ifndef HDR_hm_Default
#include "hm_Default.h"
#endif

#ifndef HDR_hm_Components
#include "hm_Components.h"
#endif

#ifndef HDR_hm_HMake
#include "hm_HMake.h"
#endif

#ifndef HDR_hm_MakeMake
#include "hm_MakeMake.h"
#endif

#ifndef HDR_hm_CMake
#include "hm_CMake.h"
#endif

// -----------------------------------------------------------------------------------------
// new-Handler
// -----------------------------------------------------------------------------------------

static void NewHandler()
{
    cerr << "Out of memory. Operation aborted." << endl;
    exit (1);
}

// -----------------------------------------------------------------------------------------
// main
// -----------------------------------------------------------------------------------------

int main (int argc, char *argv[])
{
    // NoMem-Handler installieren
    set_new_handler (NewHandler);

    hm_args Args (argc, argv);

    if (Args.HasArgs())
    {
        if (!Args.IsOk()) return 1;

        // Layout
        ifstream Stream (Args.GetLayoutPath().c_str());
        hm_layout Layout (Stream);
	if (Stream.fail() && !Stream.eof())
	    cerr << "warning: no layout-file" << endl;

        // Attributliste initialisieren
        list<hm_attribute_protocol *> Attributes;

        Attributes.push_back (new hm_default_attributes);
        // Zustzliche Attributklassen sind an dieser Stelle einzufgen

        // Spec-File
        list<string> LegalAttributes;
        for (list<hm_attribute_protocol *>::iterator Iter (Attributes.begin());
            Iter!=Attributes.end(); ++Iter)
            (*Iter)->WriteAttributes (LegalAttributes);

	int Error = 0;

        ifstream Spec (Args.GetSpecPath().c_str());
        hm_components Components (Spec, LegalAttributes);
	
        if (Components.HadErrors())
	    Error = 1;
	else if (Spec.fail() && !Spec.eof())
	{
	    cerr << "error: no spec-file" << endl;
	    Error = 1;
	}
	else
	{	
            // Make-Operatioen
            if (Args.HasOption (hm_args::HMake))
            {
                hm_hmake HMake (Layout);
                if (!HMake.Make (Components.GetComponents(), Attributes))
                    Error = 1;
            }

            if (Error==0 && Args.HasOption (hm_args::CMake))
            {
                hm_cmake CMake (Layout, Args);
                if (!CMake.Make (Components.GetComponents(), Attributes))
                    Error = 1;
            }

            if (Args.HasOption (hm_args::MakeMake))
            {
                hm_makemake MakeMake (Args);
                if (!MakeMake.Make (Components))
                    Error = 1;
            }
	}

        // Attributliste lschen
        for (list<hm_attribute_protocol *>::iterator Iter (Attributes.begin());
            Iter!=Attributes.end(); ++Iter)
            delete (*Iter);
	    
	return Error;
    }
    else
        Args.ExplainOptions (cout, "hcmake");

    return 0;
}
