//
// File:    hm_HMake.c++
// Author:  Marc Zinnschlag

#ifndef HDR_hm_HMake
#include "hm_HMake.h"
#endif

// C++

#ifndef LIB_iostream
#include <iostream>
#define LIB_iostream
#endif

#ifndef LIB_fstream
#include <fstream>
#define LIB_fstream
#endif

#ifndef LIB_algorithm
#include <algorithm>
#define LIB_algorithm
#endif

// HMake

#ifndef HDR_hm_Host
#include "hm_Host.h"
#endif

#ifndef HDR_hm_ClassMake
#include "hm_ClassMake.h"
#endif

#ifndef HDR_hm_Component
#include "hm_Component.h"
#endif

#ifndef HDR_hm_Attribute
#include "hm_Attribute.h"
#endif

// -----------------------------------------------------------------------------------------
// Klasse: hm_hmake
// -----------------------------------------------------------------------------------------

// Private Hilfsfunktionen -----------------------------------------------------------------

// Return: Ok?

bool hm_hmake::MakeComponent (const list<hm_component>& Components, hm_component& Component,
    const list<hm_attribute_protocol *>& Attributes)
{
    // Zustzliche Includes ermitteln
    list<string> Includes;
    for (list<hm_class>::const_iterator Iter (Component.GetClasses (true).begin());
        Iter!=Component.GetClasses (true).end(); ++Iter)
        for (list<hm_attribute_protocol *>::const_iterator Iter2 (
            Attributes.begin()); Iter2!=Attributes.end(); ++Iter2)
            (*Iter2)->WriteIncludes (*Iter, Includes);

    // Zustzliche Includes eintragen
    for (list<string>::const_iterator Iter (Includes.begin()); Iter!=Includes.end(); ++Iter)
        Component.AddArg (*Iter, true);

    Component.CleanIncludes();

    // Datei erzeugen
#if HOST==RISCOS
    string Path ("h.");
    Path += Component.GetName();
#else
    string Path (Component.GetName());
    Path += ".h";
#endif

    ofstream File (Path.c_str());

    d_Layout.WriteHeading ("", File, hm_layout::File);    

    File << "#ifndef HDR_" << Component.GetName() << endl
         << "#define HDR_" << Component.GetName() << endl
         << endl;

    File << "using namespace std;" << endl << endl;

    Component.WriteIncludes (File, true, Components);

    for (list<hm_class>::const_iterator Iter (Component.GetClasses (true).begin());
        Iter!=Component.GetClasses (true).end(); ++Iter)
    {
        hm_class_make ClassMake (d_Layout, *Iter, Attributes);
        ClassMake.Write (File);
    }

    File << "#endif" << endl;

    return File.good();
}

// Konstruktoren ---------------------------------------------------------------------------

hm_hmake::hm_hmake (const hm_layout& Layout) : d_Layout (Layout) {}

hm_hmake::~hm_hmake() {}

// Nicht-Konstante -------------------------------------------------------------------------

bool hm_hmake::Make (list<hm_component>& Components,
    const list<hm_attribute_protocol *>& Attributes)
{
    int Errors = 0;

    for (list<hm_component>::iterator Iter (Components.begin()); Iter!=Components.end();
        ++Iter)
        if (!(*Iter).HasMain())
        {
            if (!MakeComponent (Components, *Iter, Attributes))
            {
                cerr << "Couldn't create " << (*Iter).GetName() << ".h" << endl;
                ++Errors;
            }
        }

    if (Errors)
    {
        cerr << "Aborted with " << Errors << " error";
        if (Errors>1) cerr << 's';
        cerr << '.' << endl;
    }

    return Errors==0;
}
