//
// File:    hm_Layout.c++
// Author:  Marc Zinnschlag

#ifndef HDR_hm_Layout
#include "hm_Layout.h"
#endif

// C

#ifndef LIB_ctype
#include <cctype>
#define LIB_ctype
#endif

#ifndef LIB_stdlib
#include <cstdlib>
#define LIB_stdlib
#endif

#ifndef LIB_time
#include <ctime>
#define LIB_time
#endif

// C++

#ifndef LIB_iostream
#include <iostream>
#define LIB_iostream
#endif

#ifndef LIB_algorithm
#include <algorithm>
#define LIB_algorithm
#endif

// -----------------------------------------------------------------------------------------
// Klasse: hm_layout
// -----------------------------------------------------------------------------------------

// Statische Variablen ---------------------------------------------------------------------

static const char *s_Labels[] =
{
    "empty", // 0
    "order", // 1
    "con", // 2
    "obs", // 3
    "mani", // 4
    "stat", // 5
    "free", // 6
    "notimp", // 7
    "class", // 8
    "indent", // 9
    "fcom", // 10
    "tab", // 11
    "istyle", // 12
    "friend", // 13
    "file", // 14
//     "width", // 15
    0
};

// Private Hilfsfunktionen -----------------------------------------------------------------

void hm_layout::SetDefault()
{
    d_FunctionComment = bMode;
    d_Strict = true;
    d_EmptyState[0] = d_EmptyState[1] = d_EmptyState[2] = false;
    d_IndentWidth[0] = d_IndentWidth[1] = d_IndentWidth[3] = 4;
    d_IndentWidth[4] = 2;
    d_IndentWidth[2] = 8;
    d_TabLen = 8;
    d_IndentStyle = 0;
    d_Width = -1;
}

void hm_layout::EvalLabel (const string& Label, const string& Args)
{
    int i;
    for (i=0; s_Labels[i]; ++i)
        if (Label==s_Labels[i])
            break;

    if (s_Labels[i])
        switch (i)
        {
            case 0:

                if (Args.find ("public")!=string::npos)
                    d_EmptyState[Public] = true;

                if (Args.find ("protected")!=string::npos)
                    d_EmptyState[Protected] = true;

                if (Args.find ("private")!=string::npos)
                    d_EmptyState[Private] = true;

                break;

            case 1: d_Order = Args; break;
            case 2: d_Heading[Con] = Args; break;
            case 3: d_Heading[Obs] = Args; break;
            case 4: d_Heading[Mani] = Args; break;
            case 5: d_Heading[Stat] = Args; break;
            case 6: d_Heading[Free] = Args; break;
            case 7: d_Heading[NotImp] = Args; break;
            case 8: d_Heading[Classes] = Args; break;

            case 9:
            {
                int Len = Args.length();
                int i2 = 0;
                for (int i=0; i<5; ++i)
                {
                    while (i2<Len && !isdigit (Args[i2])) ++i2;

                    if (i2<Len)
                    {
                        d_IndentWidth[i] = atoi (Args.c_str() + i2);
                        if (d_IndentWidth[i]<0) d_IndentWidth[i] = 0;
                    }

                    while (i2<Len && isdigit (Args[i2])) ++i2;
                }

                break;
            }

            case 10:

                if (!Args.empty())
                {
                    switch (Args[0])
                    {
                        case 'a': d_FunctionComment = aMode; break;
                        case 'A': d_FunctionComment = AMode; break;
                        case 'b': d_FunctionComment = bMode; break;
                        case 'B': d_FunctionComment = BMode; break;
                    }

                    d_Strict =
                        find (Args.begin(), Args.end(), 's')!=Args.end() ||
                        find (Args.begin(), Args.end(), 'S')!=Args.end();
                }

                break;

            case 11:
            {
                d_Tab = true;

                int Len = Args.length();
                int i = 0;

                while (i<Len && !isdigit (Args[i])) ++i;

                if (i<Len)
                {
                    d_TabLen = atoi (Args.c_str() + i);
                    if (d_TabLen<0) d_TabLen = 1;
                }

                break;
            }

            case 12:
            {
                int Len = Args.length();
                int i = 0;

                while (i<Len && !isdigit (Args[i])) ++i;

                if (i<Len)
                {
                    d_IndentStyle = atoi (Args.c_str() + i);
                    if (d_IndentStyle<0 || d_IndentStyle>2) d_IndentStyle = 0;
                }

                break;
            }

            case 13: d_Heading[Friends] = Args; break;

            case 14: d_Heading[File] = Args; break;

//             case 15:
//             {
//                 int Len = Args.length();
//                 int i = 0;
//
//                 while (i<Len && !isdigit (Args[i])) ++i;
//
//                 if (i<Len)
//                 {
//                     d_Width = atoi (Args.c_str() + i);
//                     if (d_TabLen<0) d_TabLen = 1;
//                 }
//
//                 break;
//             }
        }
    else
        cerr << "Layout: Label " << Label << " unknown" << endl;
}

void hm_layout::ReadLayout (istream& Stream)
{
    char c = 0;

    do
    {
        string Label;
        string Args;

        while ((c = ReadChar (Stream)) && isalpha (c))
            Label += c;

        while (c && c!=':' && c!='\n')
            c = ReadChar (Stream);

        if (c==':')
        {
            while ((c = ReadChar (Stream)) && (c==' ' || c=='\t'));

            if (c && c!='\n')
            {
                do
                    Args += c;
                while ((c = ReadChar (Stream)) && c!='\n');
            }
        }

        int i = Args.length();
        while (i>0 && (Args[i-1]==' ' || Args[i-1]=='\t'))
            --i;

        if (!Label.empty())
            EvalLabel (Label, Args.substr (0, i));
    }
    while (c);
}

void hm_layout::MakeComplete()
{
    // Indent-Strings erzeugen
    for (int i=0; i<5; ++i)
        if (d_Tab)
        {
            int Tabs = d_IndentWidth[i]/d_TabLen;
            int Spaces = d_IndentWidth[i]%d_TabLen;

            for (int i2=0; i2<Tabs; ++i2) d_Indent[i] += '\t';
            for (int i2=0; i2<Spaces; ++i2) d_Indent[i] += ' ';
        }
        else
            for (int i2=0; i2<d_IndentWidth[i]; ++i2) d_Indent[i] += ' ';

    // Order ergnzen
    static const char *Full = "ncomsCOMSfdplt";

    for (int i=0; Full[i]; ++i)
    {
        int i2;
        int Len = d_Order.length();
        for (i2 = 0; i2<Len; ++i2)
            if (d_Order[i2]==Full[i])
                break;

        if (i2>=Len)
	{
	    if (Full[i]=='d' || Full[i]=='p' || Full[i]=='l' || Full[i]=='t')
	    {
	    	string::size_type Index =
		    d_Order.find_first_of (Full[i]=='l' ? 'm' : Full[i]-1);
		    
		if (Index==-1)
		    d_Order += Full[i];
		if (Index==d_Order.length()-1)
		    d_Order.append (1, Full[i]);
		else
		    d_Order.insert (Index+1, 1, Full[i]);
	    }
	    else	    
	    	d_Order += Full[i];
	}
    }
}

// Falls kein Zeichen mehr gelesen werden konnte, wird der Wert 0 zurckgegeben

char hm_layout::ReadChar (istream& Stream)
{
    char c;
    if (!Stream.get (c))
        return 0;

    if (c=='\\')
    {
        if (!Stream.get (c))
            return 0;

        if (c=='\n')
            return ReadChar (Stream);

        Stream.putback (c);
        return '\\';
    }
    else
        return c;
}

// Konstruktoren ---------------------------------------------------------------------------

hm_layout::hm_layout (istream& Stream) : d_Tab (false)
{
    SetDefault();
    ReadLayout (Stream);
    MakeComplete();
}

// Konstante -------------------------------------------------------------------------------

// berschrift formatiert in Stream schreiben

void hm_layout::WriteHeading (const string& Class, ostream& Stream, heading Heading) const
{
    if (!d_Heading[Heading].empty())
    {
        bool Slash = false;
        string Indent (Heading==Classes || Heading==Free || Heading==File ?
	    "" : d_Indent[Section]);

        int Len = d_Heading[Heading].length();
        Stream << Indent << "// ";
        for (int i=0; i<Len; ++i)
        {
            char c = d_Heading[Heading][i];

            if (Slash)
            {
                switch (c)
                {
                    case 'c': Stream << Class; break;
                    case 'n': Stream << endl << Indent << "// "; break;
                    case '\\': Stream << c; break;
		    
		    case 'd':
		    {
		    	const int BufferSize = 64;
		    	char Buffer[BufferSize];
			
			time_t Time = time (0);
			
			strftime (Buffer,
			
			BufferSize,
			
			"%x",
			
			localtime (&Time));
			
			Stream << Buffer;
		    
		    	break;
		    }
                }

                Slash = false;
            }
            else if (c=='\\')
                Slash = true;
            else
                Stream << c;
        }

        Stream << endl;
    }
}
