//
// File:    hm_Layout.h
// Author:  Marc Zinnschlag

// Klassen:
// hm_layout:
//      Layout-Datei

#ifndef HDR_hm_Layout
#define HDR_hm_Layout

// C++

#ifndef LIB_string2
#include <string>
#define LIB_string2
#endif

#ifndef LIB_list
#include <list>
#define LIB_list
#endif

// -----------------------------------------------------------------------------------------
// Klasse: hm_layout
// -----------------------------------------------------------------------------------------

using namespace std;

#include <iosfwd>

class hm_layout
{
    public:

        enum access
        {
            Public = 0,
            Protected = 1,
            Private = 2
        };

        enum indent
        {
            Access = 0,
            Section = 1,
            Member = 2,
            Imp = 3,
            Sub = 4
        };

        enum heading
        {
            Con = 0,
            Obs = 1,
            Mani = 2,
            Stat = 3,
            Free = 4,
            NotImp = 5,
            Classes = 6,
            Friends = 7,
	    File = 8
        };

        enum function_comment
        {
            bMode, BMode, aMode, AMode
        };

    private:

        bool d_EmptyState[3];
        function_comment d_FunctionComment;
        bool d_Strict;
        int d_IndentWidth[5];
        string d_Indent[5];
        string d_Order;
        string d_Heading[10];
        bool d_Tab;
        int d_TabLen;
        int d_IndentStyle;
        int d_Width;

    private:

    // Nicht-Konstante

        void SetDefault();

        void EvalLabel (const string& Label, const string& Args);

        void ReadLayout (istream& Stream);

        void MakeComplete();

    // Statische

        static char ReadChar (istream& Stream);
        // Falls kein Zeichen mehr gelesen werden konnte, wird der Wert 0 zurckgegeben

    public:

    // Konstruktoren

        hm_layout (istream& Stream);

    // Konstante

        bool GetEmptyState (access Access) const;
	// Abschnitt auch dann auffhren, wenn er leer ist?

        bool GetTabMode() const;
        // Verwendung von Tabs statt Spaces?

        int GetWidth() const;
        // Maximale Zahl Zeichen pro Zeile (-1: unbegrenzt)

        function_comment GetFunctionCommentMode() const;

        bool HasStrictFunctionComments() const;

        string GetIndent (indent Indent) const;

        int GetIndentWidth (indent Indent) const;

        string GetOrder() const;
        // Reihenfolge der verschiedenen Abschnitte

        string GetHeading (heading Heading) const;
        // berschrift zurckgeben (kann \-Code enthalten)

        int GetIndentStyle() const;

        void WriteHeading (const string& Class, ostream& Stream, heading Heading) const;
        // berschrift formatiert in Stream schreiben
};

// inline-Funktionen

inline bool hm_layout::GetEmptyState (access Access) const { return d_EmptyState[Access]; }
inline bool hm_layout::GetTabMode() const { return d_Tab; }
inline int hm_layout::GetWidth() const { return d_Width; }
inline hm_layout::function_comment hm_layout::GetFunctionCommentMode() const
{ return d_FunctionComment; }
inline bool hm_layout::HasStrictFunctionComments() const { return d_Strict; }
inline string hm_layout::GetIndent (indent Indent) const { return d_Indent[Indent]; }
inline int hm_layout::GetIndentWidth (indent Indent) const { return d_IndentWidth[Indent]; }
inline string hm_layout::GetOrder() const { return d_Order; }
inline string hm_layout::GetHeading (heading Heading) const { return d_Heading[Heading]; }
inline int hm_layout::GetIndentStyle() const { return d_IndentStyle; }

#endif
