//
// File:    hm_SpecScan.h
// Author:  Marc Zinnschlag

// Klassen:
// hm_spec_line:
//      Logische Zeile einer Spec-Datei; unterteilt in Anteil vor Doppelpunkt (File) und
//      nach Doppelpunkt (Args)
// hm_spec_scanner:
//      Einlesen von Spec-Datei-Zeilen

#ifndef HDR_hm_SpecScan
#define HDR_hm_SpecScan

// C++

#ifndef LIB_string2
#include <string>
#define LIB_string2
#endif

#ifndef LIB_list
#include <list>
#define LIB_list
#endif

// -----------------------------------------------------------------------------------------
// Klasse: hm_spec_line
// -----------------------------------------------------------------------------------------

using namespace std;

#include <iosfwd>

class hm_spec_line
{
    public:

        enum file_type
        {
            CFile, HFile, NoFile
        };

    private:

        string d_File;
        file_type d_Type;
        list<string> d_Args;

    public:

    // Konstruktoren

        hm_spec_line();

        hm_spec_line (const string& File, file_type Type);
        // File: Anteil links des Doppelpunkts ohne '.' und Endung

        ~hm_spec_line();

    // Konstante

        string GetFile() const;
        file_type GetType() const;
        const list<string>& GetArgs() const;

    // Nicht-Konstante

        void AddArg (const string& Arg);
        // Argument hinzufgen (Anteil rechts des Doppelpunkts)
};

// Freie Operatoren

ostream& operator<< (ostream& Stream, const hm_spec_line& Line);

// inline-Funktionen

inline string hm_spec_line::GetFile() const { return d_File; }
inline hm_spec_line::file_type hm_spec_line::GetType() const { return d_Type; }
inline const list<string>& hm_spec_line::GetArgs() const { return d_Args; }

// -----------------------------------------------------------------------------------------
// Klasse: hm_spec_scanner
// -----------------------------------------------------------------------------------------

class hm_spec_scanner
{
        istream& d_Stream;
        int d_Line;
        bool d_NewLine;
        bool d_Errors;
        list<string> d_LegalArgs;

    private:

    // Nicht implementiert

        hm_spec_scanner (const hm_spec_scanner&);
        hm_spec_scanner& operator= (const hm_spec_scanner&);

    // Nicht-Konstante

        void ReportError (const string& Error);

        void SkipLine();
        // Zum Anfang der nchsten logischen Zeile vorrcken

        bool ReadChar (char& c);
        // Ungefiltertes Zeichen lesen und ggf. Zeilennummer erhhen

        void Putback (char c);

        bool GetNext (char& c);
        // Return: Konnte Zeichen gelesen werden (sonst c undef)
        // \-Zeilenumbrche und Kommentare bereits herausgefiltert

        void SkipWhitespaces (bool SkipNewLine = false);

        bool ReadFile (string& File);
        // Return: Ok?

        bool ReadType (hm_spec_line::file_type& Type);
        // Return: Ok?

        string ReadArg();
        // Falls kein Argument mehr gelesen werden konnte, wird ein leerer String
        // zurckgegeben

    // Statische

        static bool IsWhitespace (char c);

    public:

    // Konstruktoren

        hm_spec_scanner (istream& Stream, const list<string>& LegalArgs);
        // Stream: Spec-Datei (Stream mu mindestens so lange ex. wie *this in Gebrauch ist)

        ~hm_spec_scanner();

    // Konstamte

        bool HadErrors() const;
        // Fehlerhafte Zeilen gelesen?

    // Nicht-Konstante

        bool Get (hm_spec_line& Line);
        // Logische Zeile einlesen
        // Return: Konnte Zeile eingelesen werden (sonst Line undef)
        // Falls eine fehlerhafte Zeile auftritt, wird eine Fehlermeldung nach cerr
        // geschrieben und versucht aus den nicht fehlerhaften Bestandteilen ein
        // hm_spec_line-Objekt zu generieren
        // Falls das nicht gelingt, wird die Zeile ignoriert und mit der nchsten Zeile
        // fortgefahren ohne vorher aus der Funktion zurckzukehren
};

// inline-Funktionen

inline bool hm_spec_scanner::IsWhitespace (char c) { return c==' ' || c=='\t'; }
inline bool hm_spec_scanner::HadErrors() const { return d_Errors; }

#endif
