//
// File:    le_BMatrix.c++
// Author:  Marc Zinnschlag

#ifndef HDR_le_BMatrix
#include "le_BMatrix.h"
#endif

// C

#ifndef LIB_assert
#include <cassert>
#define LIB_assert
#endif

#ifndef LIB_string
#include <cstring>
#define LIB_string
#endif

using namespace std;

// -----------------------------------------------------------------------------------------
// Klasse: le_binary_matrix
// -----------------------------------------------------------------------------------------

// Private Hilfsfunktionen -----------------------------------------------------------------

void le_binary_matrix::Init (int Size)
{
    if (Size<1) Size = 1;

    d_ArraySize = Size;
    d_Array = new char[Size*Size];

    memset (d_Array, 0, Size*Size);
}

void le_binary_matrix::Cleanup()
{
    delete [] d_Array;
}

void le_binary_matrix::Copy (const le_binary_matrix& Matrix)
{
    assert (d_ArraySize>=Matrix.d_Size);

    d_Size = Matrix.d_Size;
    for (int i=0; i<d_Size; ++i)
        memcpy (d_Array + d_ArraySize*i, Matrix.d_Array + Matrix.d_ArraySize * i, d_Size);
}


void le_binary_matrix::Warshall (bool State)
{
    for (int k=0; k<d_Size; ++k)
        for (int i=0; i<d_Size; ++i)
            if (Get (i, k))
                for (int j=0; j<d_Size; ++j)
                    if (Get (k, j))
                        Set (i, j, State);
}

// Konstruktoren ---------------------------------------------------------------------------

le_binary_matrix::le_binary_matrix (int SizeHint) : d_Size (0)
{
    if (SizeHint<=0) SizeHint = InitSize;
    Init (SizeHint);
}

// Copy-Konstruktor

le_binary_matrix::le_binary_matrix (const le_binary_matrix& Matrix)
{
    Init (Matrix.d_Size);
    Copy (Matrix);
}

le_binary_matrix::~le_binary_matrix()
{
    Cleanup();
}

// Konstante -------------------------------------------------------------------------------

bool le_binary_matrix::IsEqual (const le_binary_matrix& Matrix) const
{
    if (d_Size!=Matrix.d_Size) return false;

    for (int i=0; i<d_Size; ++i)
        if (memcmp (d_Array + i*d_ArraySize, Matrix.d_Array + i*Matrix.d_ArraySize, d_Size))
            return false;

    return true;
}

// Undef, fr ungltige Werte von Row oder Col (ungltig: <0 || >=GetSize())

bool le_binary_matrix::Get (int Row, int Col) const
{
    assert (Row>=0 && Row<d_Size);
    assert (Col>=0 && Col<d_Size);

    return d_Array[Row * d_ArraySize + Col];
}

// Nicht-Konstante -------------------------------------------------------------------------

// Zuweisung

le_binary_matrix& le_binary_matrix::operator= (const le_binary_matrix& Matrix)
{
    if (&Matrix!=this)
    {
        if (d_ArraySize<Matrix.d_Size || d_Size>Matrix.d_Size + Matrix.d_Size/2)
        {
            // Array ersetzen
            Cleanup();
            Init (Matrix.d_Size);
        }

        // Kopieren
        d_Size = Matrix.d_Size;
        for (int i=0; i<d_Size; ++i)
            memcpy (d_Array + i*d_ArraySize, Matrix.d_Array + i*Matrix.d_ArraySize, d_Size);
    }

    return *this;
}

// Undef, falls Gre der Matrizen nicht bereinstimmt

le_binary_matrix& le_binary_matrix::operator|= (const le_binary_matrix& Matrix)
{
    if (&Matrix!=this)
    {
        assert (d_Size==Matrix.d_Size);
        for (int i=0; i<d_Size; ++i)
            for (int j=0; j<d_Size; ++j)
                if (Matrix.Get (i, j)) Set (i, j);
    }

    return *this;
}

// Undef, falls Gre der Matrizen nicht bereinstimmt

le_binary_matrix& le_binary_matrix::operator&= (const le_binary_matrix& Matrix)
{
    if (&Matrix!=this)
    {
        assert (d_Size==Matrix.d_Size);
        for (int i=0; i<d_Size; ++i)
            for (int j=0; j<d_Size; ++j)
                if (!Matrix.Get (i, j)) Clear (i, j);
    }

    return *this;
}

// Undef, falls Gre der Matrizen nicht bereinstimmt

le_binary_matrix& le_binary_matrix::operator^= (const le_binary_matrix& Matrix)
{
    if (&Matrix==this)
        Clear();
    else
    {
        assert (d_Size==Matrix.d_Size);
        for (int i=0; i<d_Size; ++i)
            for (int j=0; j<d_Size; ++j)
                Set (i, j, Get (i, j) ^ Matrix.Get (i, j));
    }

    return *this;
}

// Undef, fr ungltige Werte von Row oder Col (ungltig: <0 || >=GetSize())

void le_binary_matrix::Set (int Row, int Col, bool State)
{
    assert (Row>=0 && Row<d_Size);
    assert (Col>=0 && Col<d_Size);

    d_Array[Col + Row*d_ArraySize] = State;
}

// Undef, fr ungltige Werte von Row oder Col (ungltig: <0 || >=GetSize())

void le_binary_matrix::Clear (int Row, int Col)
{
    assert (Row>=0 && Row<d_Size);
    assert (Col>=0 && Col<d_Size);

    Set (Row, Col, false);
}

// Alle Eintrge setzen

void le_binary_matrix::Set()
{
    for (int i=0; i<d_Size; ++i)
        memset (d_Array + i*d_ArraySize, 1, d_Size);
}

// Alle Eintrge lschen

void le_binary_matrix::Clear()
{
    for (int i=0; i<d_Size; ++i)
        memset (d_Array + i*d_ArraySize, 0, d_Size);
}

// Neue Zeile und Spalte erzeugen und Index zurckgeben

int le_binary_matrix::Add()
{
    if (d_Size>=d_ArraySize)
    {
        // Array ersetzen
        int OldArraySize = d_ArraySize;
        char *OldArray = d_Array;

        Init (d_ArraySize * GrowFactor);

        // Kopieren
        for (int i=0; i<d_Size; ++i)
            memcpy (d_Array + i*d_ArraySize, OldArray + i*OldArraySize, d_Size);

        // Altes Array lschen
        delete [] OldArray;
    }

    return d_Size++;
}

// Redundante Kanten einfgen

void le_binary_matrix::MakeTransitive()
{
    Warshall (true);
}

// Redundante Kanten entfernen
// Undef, falls Matrix nicht alle redundanten Kanten eingefgt waren

void le_binary_matrix::MakeNonTransitive()
{
    Warshall (false);

    // Einheitsmatrix abziehen
    for (int i=0; i<d_Size; ++i)
        Clear (i, i);
}
