//
// File:    le_BMatrix.h
// Author:  Marc Zinnschlag

// Klassen:
// le_binary_matrix:
//      Quadratische Matrix mit binren Eintrgen

#ifndef HDR_le_BMatrix
#define HDR_le_BMatrix

// -----------------------------------------------------------------------------------------
// Klasse: le_binary_matrix
// -----------------------------------------------------------------------------------------

class le_binary_matrix
{
        char *d_Array;
        int d_Size; // Lnge einer Spalte
        int d_ArraySize; // Lnge einer Array-Spalte (totale Lnge: d_ArraySize*d_ArraySize)

        enum { InitSize = 8, GrowFactor = 2 };

    private:

    // Nicht-Konstante

        void Init (int Size);

        void Cleanup();

        void Copy (const le_binary_matrix& Matrix);

        void Warshall (bool State);

    public:

    // Konstruktoren

        le_binary_matrix (int SizeHint = -1);

        le_binary_matrix (const le_binary_matrix& Matrix);
        // Copy-Konstruktor

        ~le_binary_matrix();

    // Konstante

        bool IsEqual (const le_binary_matrix& Matrix) const;

        int GetSize() const;

        bool Get (int Row, int Col) const;
        // Undef, fr ungltige Werte von Row oder Col (ungltig: <0 || >=GetSize())

    // Nicht-Konstante

        le_binary_matrix& operator= (const le_binary_matrix& Matrix);
        // Zuweisung

        le_binary_matrix& operator|= (const le_binary_matrix& Matrix);
        // Undef, falls Gre der Matrizen nicht bereinstimmt

        le_binary_matrix& operator&= (const le_binary_matrix& Matrix);
        // Undef, falls Gre der Matrizen nicht bereinstimmt

        le_binary_matrix& operator^= (const le_binary_matrix& Matrix);
        // Undef, falls Gre der Matrizen nicht bereinstimmt

        void Set (int Row, int Col, bool State = true);
        // Undef, fr ungltige Werte von Row oder Col (ungltig: <0 || >=GetSize())

        void Clear (int Row, int Col);
        // Undef, fr ungltige Werte von Row oder Col (ungltig: <0 || >=GetSize())

        void Set();
        // Alle Eintrge setzen

        void Clear();
        // Alle Eintrge lschen

        int Add();
        // Neue Zeile und Spalte erzeugen und Index zurckgeben

        void MakeTransitive();
	// Redundante Kanten einfgen

        void MakeNonTransitive();
        // Redundante Kanten entfernen
        // Undef, falls Matrix nicht alle redundanten Kanten eingefgt waren
};

// Freie Operatoren

bool operator== (const le_binary_matrix& Left, const le_binary_matrix& Right);
bool operator!= (const le_binary_matrix& Left, const le_binary_matrix& Right);

// inline-Funktionen

inline int le_binary_matrix::GetSize() const { return d_Size; }
inline bool operator== (const le_binary_matrix& Left, const le_binary_matrix& Right)
{ return Left.IsEqual (Right); }
inline bool operator!= (const le_binary_matrix& Left, const le_binary_matrix& Right)
{ return !(Left==Right); }

#endif
