//
// File:    ul_Scanner.h
// Author:  Marc Zinnschlag

// Klassen:
// ul_scanner:
//      Scanner fr C++-Quelltext

#ifndef HDR_ul_Scanner
#define HDR_ul_Scanner

#ifndef HDR_ul_Token
#include "ul_Token.h"
#endif

// -----------------------------------------------------------------------------------------
// Klasse: ul_scanner
// -----------------------------------------------------------------------------------------

using namespace std;

#include <iosfwd>

class ul_token;
class ul_putback_buffer;

class ul_scanner
{
        istream& d_Stream;
        ul_putback_buffer *d_PutbackBuffer;
        bool d_EOF;
        int d_State;
        int d_FilterFlags;
        bool d_IgnorePreprocessor;

        enum
        {
            WhitespaceState,
            SlashState,
            CppCommentState,
            CCommentState,
            CCommentEndState,
            CharState, // Literal
            CharBSState, // Backslash
            CharEndState, // ' erwartet
            CharGarbageState,
            StringState,
            StringBSState, // Backslash
            NumberState,
            ZeroState, // 0 am Anfang einer Zahl
            PointState, // Dezimalpunkt am Anfang einer Zahl
            NumberEState,
            XState,
            Number2State, // Nach Dezimalpunkt
            Number3State,  // Nach e
            NameState,
            SpecialState
        };

        enum { Endmark = 0x1000 };

    private:

    // Nicht implementiert

        ul_scanner (const ul_scanner&);
        ul_scanner& operator= (const ul_scanner&);

    // Konstante

        bool IsToBeIgnored (const ul_token& Token) const;

    // Nicht-Konstante

        void MakeNewToken (ul_token *Token);

        int GetChar();
        // Falls Dateiende erreicht, wird Endmark zurckgegeben

        int InterpretChar (int Char, ul_token *Token, string& String);
        // Return:
        // -1 Token unvollstndig, Char nicht an String anhngen
        //  0 Token unvollstndig
        //  1 Token vollstndig, Sprung zu WhitespaceState
        //  2 Token vollstndig, letztes Zeichen zurcklegen, Sprung zu WhitespaceState

    public:

    // Konstruktoren

        ul_scanner (istream& Stream, int PutbackSizeHint = 4,
            bool IgnorePreprocessor = true);

        ~ul_scanner();

    // Konstante

        int GetPutbackSize() const;
        // Anzahl Tokens in Putback-Buffer zurckgeben

        bool IsEOF() const;
        // EOF-Status zurckgeben

        operator const void *() const;
        // Ende des Files noch nicht erreicht?

    // Nicht-Konstante

        int GetNext (ul_token *Token);
        // Nchstes Token lesen
        // Falls ein EOF-Token zurckgegeben wird, wird fr alle weiteren Aufrufe
        // dieser Funktion dasselbe Token zurckgegeben
        // Return: 0 EOF-Token erreicht, 1 anderes Token

        void Putback (const ul_token& Token);
        // Token zurckliegen (mu nicht unbedingt dem aus GetNext enthaltenen Token
        // entsprechend)

        void Enable (ul_token::type TokenType);
        // Angegebenen TokenType zulassen
        // Alle Types bei Konstruktion zugelassen

        void Disable (ul_token::type TokenType);
        // Angegebenen TokenType unterdrcken
        // Tokens der Typen EndOfFile und Empty knnen nicht unterdrckt werden (Aufruf von
        // Disable fr diese Typen wird ignoriert)

        void EnablePreprocessor();
        // Preprocessor-Anweisungen durchlassen

        void DisablePreprocessor();
        // Preprocessor-Anweisungen ignorieren
};

// Freie Operatoren

ul_scanner& operator<< (ul_scanner& Scanner, const ul_token& Token);
ul_scanner& operator>> (ul_scanner& Scanner, ul_token& Token);

// inline-Funktionen

inline bool ul_scanner::IsEOF() const { return d_EOF; }
inline void ul_scanner::EnablePreprocessor() { d_IgnorePreprocessor = false; }
inline void ul_scanner::DisablePreprocessor() { d_IgnorePreprocessor = true; }

#endif
