//
// File:    ul_Token.c++
// Author:  Marc Zinnschlag

#ifndef HDR_ul_Token
#include "ul_Token.h"
#endif

// C

#ifndef LIB_assert
#include <cassert>
#define LIB_assert
#endif

#ifndef LIB_string
#include <cstring>
#define LIB_string
#endif

// C++

#ifndef LIB_iostream
#include <iostream>
#define LIB_iostream
#endif

// -----------------------------------------------------------------------------------------
// Klasse: ul_token
// -----------------------------------------------------------------------------------------

// Statische Variablen ---------------------------------------------------------------------

static const char *s_SpecialTable[ul_token::SpecialIndices] =
{
    "++", "--", "==", "!=", "<=", ">=", "::", "&&", "||", "<<", ">>"
};

// Private Hilfsfunktionen -----------------------------------------------------------------

ul_token::data_type ul_token::GetDataType() const
{
    switch (d_Type)
    {
        case Name:
        case String:
        case Number:
        case CComment:
        case CppComment:
        case GarbageString:
        case GarbageCharLiteral:
        case GarbageNumber:

            return StringData;

        case CharLiteral:
        case Char:
        case Special:
        case Whitespace:

            return CharData;

        default:

            return NoData;
    }
}

void ul_token::MakeString (const string& String)
{
    d_Data.d_String = new char[String.length()+1];
    strcpy (d_Data.d_String, String.c_str());
}

void ul_token::Copy (const ul_token& Token)
{
    d_Type = Token.d_Type;
    switch (GetDataType())
    {
        case StringData:

            MakeString (Token.d_Data.d_String);
            break;

        case CharData:

            d_Data.d_Char = Token.d_Data.d_Char;
            break;
    }
}

void ul_token::Cleanup()
{
    if (GetDataType()==StringData)
        delete [] d_Data.d_String;
}

// Konstruktoren ---------------------------------------------------------------------------

// Leeres Token oder EndOfFile-Token

ul_token::ul_token (type Type) : d_Type (Type)
{
    assert (Type==Empty || Type==EndOfFile);
}

// Name, String, Number, CComment, CppComment, GarbageString, GarbageCharLiteral
// oder GarbageNumber Token (undef fr alle anderen Typen)

ul_token::ul_token (type Type, const string& String) : d_Type (Type)
{
    assert (GetDataType()==StringData);
    MakeString (String);
}

// CharLiteral, Char, Special, Whitespace (undef fr alle anderen Typen)

ul_token::ul_token (type Type, char c) : d_Type (Type)
{
    assert (GetDataType()==CharData);
    assert (Type!=Special || c<SpecialIndices);

    d_Data.d_Char = c;
}

// Copy-Konstruktor

ul_token::ul_token (const ul_token& Token)
{
    Copy (Token);
}

ul_token::~ul_token()
{
    Cleanup();
}

// Konstante -------------------------------------------------------------------------------

// Name, String, Number, CComment, CppComment, GarbageString, GarbageCharLiteral
// oder GarbageNumber Token (undef fr alle anderen Typen)

string ul_token::GetString() const
{
    assert (GetDataType()==StringData);
    return d_Data.d_String;
}

// CharLiteral, Char, Special, Whitespace (undef fr alle anderen Typen)

char ul_token::GetChar() const
{
    assert (GetDataType()==CharData);
    return d_Data.d_Char;
}

// Undef, falls Token nicht vom Type Special ist

string ul_token::GetSpecialString() const
{
    assert (d_Type==Special);
    return GetSpecialString (d_Data.d_Char);
}

// Nicht-Konstante -------------------------------------------------------------------------

// Zuweisung

ul_token& ul_token::operator= (const ul_token& Token)
{
    if (&Token!=this)
    {
        Cleanup();
        Copy (Token);
    }

    return *this;
}

// Statische -------------------------------------------------------------------------------

string ul_token::GetSpecialString (int Index)
{
    assert (Index>=0 && Index<SpecialIndices);
    return s_SpecialTable[Index];
}

// Freie Operatoren ------------------------------------------------------------------------

int operator== (const ul_token& Left, const ul_token& Right)
{
    if (Left.GetType()!=Right.GetType()) return 0;
    switch (Left.GetDataType())
    {
        case Left.StringData:

            return Left.GetString()==Right.GetString();

        case Left.CharData:

            return Left.GetChar()==Right.GetChar();

        default:

            return 1;
    }
}

// Token-Text ohne Typ-Angabe in Stream schreiben
// Bei Tokenerzeugung entfernte Zeichen (z.B. " " bei Strings) werden wieder ergnzt

ostream& operator<< (ostream& Stream, const ul_token& Token)
{
    switch (Token.GetDataType())
    {
        case Token.StringData:
        {
            switch (Token.GetType())
            {
                case Token.String: Stream << '"' << Token.GetString() << '"'; break;
                case Token.CComment: Stream << "/*" << Token.GetString() << "*/"; break;
                case Token.CppComment: Stream << "//" << Token.GetString(); break;
                default: Stream << Token.GetString();
            }

            break;
        }

        case Token.CharData:
        {
            switch (Token.GetType())
            {
                case Token.CharLiteral:
                {
                    if (Token.GetChar()=='\\' || Token.GetChar()=='\'')
                        Stream << "'\\" << Token.GetChar() << '\'';
                    else
                        Stream << '\'' << Token.GetChar() << '\'';

                    break;
                }

                case Token.Special: Stream << Token.GetSpecialString(); break;
                default: Stream << Token.GetChar();
            }

            break;
        }
    }

    return Stream;
}
