//
// File:    ul_Token.h
// Author:  Marc Zinnschlag

// Klassen:
// ul_token:
//      Token fr C++-Programmtexte

#ifndef HDR_ul_Token
#define HDR_ul_Token

// C++

#ifndef LIB_string2
#include <string>
#define LIB_string2
#endif

// -----------------------------------------------------------------------------------------
// Klasse: ul_token
// -----------------------------------------------------------------------------------------

using namespace std;

#include <iosfwd>

class ul_token
{
    public:

        enum type
        {
            Name = 0,
            String = 1, // Ohne " "
            CharLiteral = 2, // Ohne ' '
            Char = 3,
            Special = 4,
            Number = 5,
            CComment = 6, // ohne /* und */
            CppComment = 7, // ohne //
            Whitespace = 8,
            EndOfFile = 9,
            GarbageString = 10, // Einschlielich " " (sofern vorhanden)
            GarbageCharLiteral = 11, // Einschlielich ' ' (sofern vorhanden)
            GarbageNumber = 12,
            Empty = 13
        };

    private:

        type d_Type;

        union
        {
            char d_Char;
            char *d_String;
        } d_Data;

        enum data_type
        {
            NoData,
            StringData,
            CharData
        };

    public:

        enum // SpecialIndex
        {
            Inc = 0, // ++
            Dec = 1, // --
            Eq = 2, // ==
            NEq = 3, // !=
            SmTh = 4, // <=
            LaTh = 5, // >=
            Scope = 6, // ::
            LogAnd = 7, // &&
            LogOr = 8, // ||
            LeftShift = 9, // <<
            RightShift = 10, // >>

            SpecialIndices = 11
        };

    private:

    // Konstante

        data_type GetDataType() const;

    // Nicht-Konstante

        void MakeString (const string& String);

        void Copy (const ul_token& Token);

        void Cleanup();

    public:

    // Konstruktoren

        ul_token (type Type = Empty);
        // Leeres Token oder EndOfFile-Token

        ul_token (type Type, const string& String);
        // Name, String, Number, CComment, CppComment, GarbageString, GarbageCharLiteral
        // oder GarbageNumber Token (undef fr alle anderen Typen)

        ul_token (type Type, char c);
        // CharLiteral, Char, Special, Whitespace (undef fr alle anderen Typen)

        ul_token (const ul_token& Token);
        // Copy-Konstruktor

        ~ul_token();

    // Konstante

        type GetType() const;

        string GetString() const;
        // Name, String, Number, CComment, CppComment, GarbageString, GarbageCharLiteral
        // oder GarbageNumber Token (undef fr alle anderen Typen)

        char GetChar() const;
        // CharLiteral, Char, Special, Whitespace (undef fr alle anderen Typen)

        string GetSpecialString() const;
        // Undef, falls Token nicht vom Type Special ist

    // Nicht-Konstante

        ul_token& operator= (const ul_token& Token);
        // Zuweisung

    // Statische

        static string GetSpecialString (int Index);

    friend int operator== (const ul_token& Left, const ul_token& Right);
    friend ostream& operator<< (ostream& Stream, const ul_token& Token);
};

// Freie Operatoren

int operator== (const ul_token& Left, const ul_token& Right);
int operator!= (const ul_token& Left, const ul_token& Right);

ostream& operator<< (ostream& Stream, const ul_token& Token);
// Token-Text ohne Typ-Angabe in Stream schreiben
// Bei Tokenerzeugung entfernte Zeichen (z.B. " " bei Strings) werden wieder ergnzt

// inline-Funktionen

inline ul_token::type ul_token::GetType() const { return d_Type; }
inline int operator!= (const ul_token& Left, const ul_token& Right)
{ return !(Left==Right); }

#endif
